"""
Functions to generate MeasurementSets. The motivation is to generate on-the-fly MSs in
unit tests that can be used in pytest fixtures or external projects that import
xradio.testing.

All functions that depend on casacore have been moved to this module.
"""

from contextlib import contextmanager
import datetime
import itertools
from pathlib import Path
from typing import Generator, Tuple, Dict, Any, Optional, Iterable

import numpy as np

import casacore.tables as tables
from casacore.tables import default_ms, default_ms_subtable
from casacore.tables.tableutil import makedminfo, maketabdesc
from casacore.tables.msutil import complete_ms_desc, makearrcoldesc, required_ms_desc

# 2 observations, 2 fields, 2 states
# 2 SPWs, 4 polarizations
default_ms_descr = {
    "nrows_per_ddi": 300,
    "nchans": 16,
    "npols": 2,
    "data_cols": ["DATA"],  # ['CORRECTED_DATA'],
    # DATA / CORRECTED, etc.
    # subtables needed to test xds structure
    "SPECTRAL_WINDOW": {"0": 0, "1": 1},
    "POLARIZATION": {"0": 0, "1": 1},  # "2": 2, "3": 3},
    "ANTENNA": {"0": 0, "1": 1, "2": 2, "3": 3, "4": 4},
    # subtables neded to test partitioning
    "FIELD": {"0": 0, "1": 1},
    "EPHEMERIDES": {0: 0},
    "SCAN": {
        "1": {
            "0": {"intent": "CAL_ATMOSPHERE#ON_SOURCE"},
            "1": {"intent": "CAL_ATMOSPHERE#OFF_SOURCE"},
        },
        "2": {
            "0": {"intent": "OBSERVE_TARGET#ON_SOURCE"},
            "1": {"intent": "OBSERVE_TARGET#ON_SOURCE"},
            "2": {"intent": "OBSERVE_TARGET#ON_SOURCE"},
            "3": {"intent": "OBSERVE_TARGET#ON_SOURCE"},
        },
        "3": {
            "0": {"intent": "CALIBRATE_DELAY#ON_SOURCE,CALIBRATE_PHASE#ON_SOURCE"},
            "1": {"intent": "CALIBRATE_DELAY#ON_SOURCE,CALIBRATE_PHASE#ON_SOURCE"},
            "2": {"intent": "CALIBRATE_DELAY#ON_SOURCE,CALIBRATE_PHASE#ON_SOURCE"},
        },
    },
    # CALIBRATE_ATMOSPHERE#OFF_SOURCE,CALIBRATE_ATMOSPHERE#ON_SOURCE,CALIBRATE_WVR#OFF_SOURCE,CALIBRATE_WVR#ON_SOURCE
    "STATE": {
        "0": {"id": 0, "intent": "CAL_ATMOSPHERE#ON_SOURCE"},
        "1": {"id": 1, "intent": "CAL_ATMOSPHERE#OFF_SOURCE"},
    },
    "OBSERVATION": {"0": 0, "1": 1},
    # Mandatory as per MSv2 but not essential to be able to build xdss
    "FEED": {"0": 0},
    "POINTING": {},
    "PROCESSOR": {"0": 0, "1": 1},
    "SOURCE": {"0": 0, "1": 1},
    # Auto-generated without parameters for now:
    # 'FLAG_CMD': {},
    # 'HISTORY': {},
}


def _gen_test_ms_impl(
    msname: str,
    descr: dict = None,
    opt_tables: bool = True,
    vlbi_tables: bool = True,
    required_only: bool = True,
    misbehave: bool = False,
):
    """
    Generates an MS for testing purposes, including main table and
    subtables, either only the required ones or all. Follows the MSv2
    definitions. The aim is to produce MSs with the minimal structure and
    set of rows to exercise xradio functions.

    - To be able to effectively start testing some functions, these are
    required: MAIN, DATA_DESCRIPTION, SPECTRAL_WINDOW, POLARIZATION, ANTENNA
    - To be able to start testing partitioning: FIELD, STATE
    - Furhter partitioning testing: observation, processor, feed?

    - The optional tables SYSCAL, WEATHER are generated with very (too)
      simple contents

    Parameters
    ----------
    msname : str
        name of MS on disk
    descr : dict (Default value = None)
        MS description, including description of scans, fields,
        intents, etc. By default (empty dict) will create an MS with minimal
        structure (one observation, one array, one scan, one field, etc.)
    opt_tables : bool (Default value = True)
        whether to produce optional (sub)tables, such as SOURCE, WEATHER
    vlbi_tables : bool (Default value = True)
        whether to add optional (sub)tables GAIN_CURVE and PHASE_CAL
    required_only : bool (Default value = True)
        whether to use the complete or required columns spec
    misbehave : bool (Default value = False)
        whether to generate a misbehaving MS. For example, usual or more
        corner case conformance issues such as absence of STATE subtable,
        missing FEED subtable, presence of missing SOURCE_IDs in the FIELD
        subtable, no ASDM_EXECBLOCK table, no PROCESSOR subtable, etc.
        Additional, more minor misbehaviors:
        - irregular INTERVAL in main table, empty SPW names.
        - missing metadata for MAIN/INTERVAL, SPECTRAL_WINDOW/CHAN_WIDTH
        - missing posrefsys in EPHEM metadata

    Returns
    -------

    """
    if not descr:
        descr = default_ms_descr

    # Definitely needed: main table + following subtables: DATA_DESCRIPTPION
    #  + SPECTRAL_WINDOW + POLARIZATION
    #  + ANTENNNA (just for the names)
    # All these are required to create the xdss (they define data and dims)
    outdescr = gen_main_table(msname, descr, required_only, misbehave)
    gen_subt_ddi(msname, descr["SPECTRAL_WINDOW"], descr["POLARIZATION"])
    gen_subt_spw(msname, descr["SPECTRAL_WINDOW"], descr["nchans"], misbehave=misbehave)
    gen_subt_antenna(msname, descr["ANTENNA"])
    gen_subt_pol_setup(msname, descr["npols"], descr["POLARIZATION"])

    gen_ephem = not misbehave
    # Also needed for partitioning: FIELD, STATE
    gen_subt_field(msname, descr["FIELD"], gen_ephem=gen_ephem, misbehave=misbehave)
    if not misbehave:
        gen_subt_state(msname, descr["STATE"])

    # Required by MSv2/v3 but not strictly required to load and partition:

    # BEAM (only MSv3)

    # FEED
    gen_subt_feed(
        msname, descr["FEED"], descr["ANTENNA"], descr["SPECTRAL_WINDOW"], misbehave
    )

    # HISTORY
    gen_subt_history(msname, descr["OBSERVATION"])

    # OBSERVATION
    gen_subt_observation(msname, descr["OBSERVATION"])

    # POINTING: TODO
    gen_subt_pointing(msname)

    # PROCESSOR
    if not misbehave:
        gen_subt_processor(msname, descr["PROCESSOR"])

    if opt_tables:
        # DOPPLER: no examples seen in test MSs

        # EPHEMERIDES (only defined in Msv3, never seen in practice)

        # FLAG_CMD (made optional in MSv3, required in MSv2)
        gen_subt_flag_cmd(msname)

        # FREQ_OFFSET: no examples seen in test MSs

        # INTERFEROMETER_MODEL (only MSv3)

        # PHASED_ARRAY (only MSv3)

        # QUALITY_FREQUENCY_STATISTIC (only MSv3: listend but never defined)

        # QUALITY_BASELINE_STATISTIC (only MSv3: listend but never defined)

        # QUALITY_TIME_STATISTIC (only MSv3: listend but never defined)

        # SOURCE
        gen_subt_source(msname, descr["SOURCE"], misbehave)

        # SCAN (Only MSv3)

        # SYSCAL
        gen_subt_syscal(msname, descr["ANTENNA"])

        # WEATHER
        gen_subt_weather(msname)

        # ASDM_* subtables. One simple example
        gen_subt_asdm_receiver(msname)

        if not misbehave:
            # ASDM_EXECBLOCK is used in create_info_dicts when available
            gen_subt_asdm_execblock(msname)
            # ASDM_STATION is used in create_weather
            gen_subt_asdm_station(msname)

    if vlbi_tables:
        gen_subt_gain_curve(msname, descr["ANTENNA"])
        gen_subt_phase_cal(msname, descr["ANTENNA"])

    outdescr["params"] = {
        "opt_tables": opt_tables,
        "vlbi_tables": vlbi_tables,
        "misbehave": misbehave,
    }

    return outdescr


def gen_test_ms(
    msname: str,
    *,
    descr: Optional[Dict[str, Any]] = None,
    opt_tables: bool = True,
    vlbi_tables: bool = True,
    required_only: bool = True,
    misbehave: bool = False,
) -> Tuple[str, Dict[str, Any]]:
    """
    Generate an MS on disk and return both the path and the generated description.

    This is the high-level helper used by tests and benchmarks. The lower-level
    implementation is `_gen_test_ms_impl`.
    """

    ms_descr = _gen_test_ms_impl(
        msname,
        descr,
        opt_tables=opt_tables,
        vlbi_tables=vlbi_tables,
        required_only=required_only,
        misbehave=misbehave,
    )
    return msname, ms_descr


def gen_minimal_ms(
    msname: str = "test_msv2_minimal_required.ms",
    *,
    misbehave: bool = False,
    include_optional_tables: bool = True,
    include_vlbi_tables: bool = True,
) -> Tuple[str, Dict[str, Any]]:
    """
    Generate a minimal MSv2 dataset with sensible defaults used across tests/benchmarks.
    """

    return gen_test_ms(
        msname,
        opt_tables=include_optional_tables,
        vlbi_tables=include_vlbi_tables,
        required_only=True,
        misbehave=misbehave,
    )


def make_ms_empty(name: str, descr: dict = None, complete: bool = False):
    """
    OLD simple function. makes empty (0 rows) MSs

    Parameters
    ----------
    name : str

    descr : dict (Default value = None)

    complete : bool (Default value = False)

    Returns
    -------

    """
    if complete:
        tabdesc = complete_ms_desc("MAIN")
    else:
        tabdesc = required_ms_desc("MAIN")

    datacoldesc = tables.makearrcoldesc(
        "DATA",
        0.1 + 0.1j,
        valuetype="complex",
        ndim=2,
        datamanagertype="TiledShapeStMan",
        datamanagergroup="TiledData",
        comment="The data column",
    )
    del datacoldesc["desc"]["shape"]
    tabdesc.update(tables.maketabdesc(datacoldesc))

    weightspeccoldesc = tables.makearrcoldesc(
        "WEIGHT_SPECTRUM",
        1.0,
        valuetype="float",
        ndim=2,
        datamanagertype="TiledShapeStMan",
        datamanagergroup="TiledWgtSpectrum",
        comment="Weight for each data point",
    )
    del weightspeccoldesc["desc"]["shape"]
    tabdesc.update(tables.maketabdesc(weightspeccoldesc))

    vis = tables.default_ms(name, tabdesc=tabdesc, dminfo=makedminfo(tabdesc))
    assert vis.nrows() == 0


def gen_main_table(
    mspath: str, descr: dict, required_only: bool = True, misbehave: bool = False
):
    """
    Create main MSv2 table.
    Relies on the required/complete_ms_desc descriptions of columns

    Simplifications/assumptions:
    - All polarization setups have the same number of correlations/
    - All scans have the same SPWs (all SPWs)
    - Multiple observation: just repeat the same structure (same scans,
      fields, etc.)
    - Can generate multiple *_DATA columns but they all have the same data
      pattern

    :return: outdescr

    Parameters
    ----------
    mspath: str :

    descr: dict :

    required_only: bool :
         (Default value = True)

    misbehave : bool (Default value = False)
        all STATE_ID values are set to -1 (assuming a "misbehaved" MS with empty STATE subtable)

    Returns
    -------

    """

    outdescr = descr.copy()

    if descr == {}:
        nchans = 16
        npols = 2
    else:
        nchans = descr["nchans"]
        npols = descr["npols"]

    if required_only:
        ms_desc = required_ms_desc("MAIN")
    else:
        ms_desc = complete_ms_desc("MAIN")

    ms_desc["UVW"].update(
        options=0,
        shape=[3],
        ndim=1,
        dataManagerGroup="UVWGroup",
        dataManagerType="TiledColumnStMan",
    )
    dmgroups_spec = {"UVW": {"DEFAULTTILESHAPE": [3, nchans]}}

    # data columns. TODO: move to function - make_data_cols_descr
    for data_col_name in descr["data_cols"]:
        data_col_desc = makearrcoldesc(
            data_col_name,
            0.0,
            options=4,
            valuetype="complex",
            # Beware of the additional column description entry - not present in casatestdata
            shape=[nchans, npols],
            ndim=2,
            datamanagertype="TiledColumnStMan",
            datamanagergroup="DataGroup",
            comment="added by gen_test_ms",
        )
        ms_desc.update(maketabdesc(data_col_desc))

        dmgroups_spec.update(
            {"DataColsGroup": {"DEFAULTTILESHAPE": [nchans, npols, nchans]}}
        )
        ms_data_man_info = makedminfo(ms_desc, dmgroups_spec)

    if "STATE" not in descr:
        vis = tables.default_ms(mspath, tabdesc=ms_desc, dminfo=makedminfo(ms_desc))
        assert vis.nrows() == 0
        return
    # else:
    #    populate (TODO)

    # Key columns (that define data rows):
    # - TIME: given in descr
    # - ANTENNA1, ANTENNA2: given in descr
    # - DATA_DESC_ID, PROCESSOR_ID: given in descr
    # - FEED1, FEED2, assumed 0
    # - Not considered: ANTENNA3, FEED3, PHASE_ID, TIME_EXTRA_PREC

    with default_ms(mspath, ms_desc, ms_data_man_info) as msv2:
        desc = msv2.getcoldesc("UVW")
        assert desc["dataManagerType"] == "TiledColumnStMan"
        dminfo = msv2.getdminfo("UVW")
        assert dminfo["NAME"] == "UVWGroup"

        # Figure out amount of rows and related IDs
        nrows = descr["nrows_per_ddi"]
        dd_pairs = list(
            itertools.product(
                list(descr["SPECTRAL_WINDOW"].values()), descr["POLARIZATION"].values()
            )
        )
        nddis = len(dd_pairs)
        outdescr["nddis"] = nddis
        dd_ids = np.arange(nddis)
        nrows *= nddis
        # problem: TIME - SPW

        msv2.addrows(nrows)
        # trasnposing so that indices increase with channel number, not pol
        vis_val = np.arange(nchans * npols).reshape(
            (npols, nchans)
        ).transpose() * np.complex64(1 - 1j)
        # vis_val *= np.arange(
        # msv2.putcell("CORRECTED_DATA", 0, vis_val)
        for data_col in descr["data_cols"]:
            msv2.putcol(data_col, np.broadcast_to(vis_val, (nrows, nchans, npols)))

        # === Key attributes ===
        # Make Ids

        # TIME
        CASACORE_TO_DATETIME_CORRECTION = 3_506_716_800.0
        start = (
            datetime.datetime(
                2025, 5, 1, 1, 1, tzinfo=datetime.timezone.utc
            ).timestamp()
            + CASACORE_TO_DATETIME_CORRECTION
        )
        time_col = np.arange(nrows) + start
        msv2.putcol("TIME", time_col)

        # (TIME_EXTRA_PREC): nothing for now

        nants = len(descr["ANTENNA"])
        # ANTENNA1, ANTENNA2
        # baselines = list(itertools.product(ants, ants))
        # combinations without repetitions: baselines as list of tuples
        ants = np.arange(nants)
        baselines = list(itertools.combinations(ants, 2))
        ant1_ant2 = list(zip(*baselines))

        # TODO: needs fixes for rounding
        _reps = np.tile(ant1_ant2[0], int(nrows / len(baselines)))
        msv2.putcol("ANTENNA1", np.tile(ant1_ant2[0], int(nrows / len(baselines))))
        msv2.putcol("ANTENNA2", np.tile(ant1_ant2[1], int(nrows / len(baselines))))

        # (ANTENNA3): nothing for now

        msv2.putcol("FEED1", np.broadcast_to(0, (nrows)))

        msv2.putcol("FEED2", np.broadcast_to(0, (nrows)))

        # (FEED3): nothing for now

        # DATA_DESC_ID
        # msv2.putcol("DATA_DESC_ID", np.broadcast_to(0, (nrows)))
        ddi_col = np.repeat(dd_ids, nrows / nddis)
        msv2.putcol("DATA_DESC_ID", np.broadcast_to(ddi_col, (nrows)))

        msv2.putcol("PROCESSOR_ID", np.broadcast_to(0, (nrows)))

        msv2.putcol("FIELD_ID", np.broadcast_to(0, (nrows)))

        # PHASE_ID: nothing for now

        # === Non-key attributes ===

        # The ones that are IDs
        msv2.putcol("OBSERVATION_ID", np.broadcast_to(0, (nrows)))

        msv2.putcol("ARRAY_ID", np.broadcast_to(0, (nrows)))

        msv2.putcol("SCAN_NUMBER", np.broadcast_to(1, (nrows)))

        # STATE_ID: if no states/intents => all STATE_ID = -1
        if misbehave:
            msv2.putcol("STATE_ID", np.broadcast_to(-1, (nrows)))
        else:
            msv2.putcol("STATE_ID", np.broadcast_to(0, (nrows)))

        # Make other scalar / inc columns

        interval = 1.0
        msv2.putcol("INTERVAL", np.broadcast_to(interval, (nrows)))
        if misbehave:
            msv2.putcell("INTERVAL", 0, interval - 0.1)
            msv2.removecolkeyword("INTERVAL", "QuantumUnits")

        exposure = 0.9 * interval
        msv2.putcol("EXPOSURE", np.broadcast_to(exposure, (nrows)))

        # TIME_CENTROID
        msv2.putcol("TIME_CENTROID", time_col)

        # (PULSAR_BIN)

        # (PULSAR_GATE_ID)

        # (BASELINE_REF)

        # UVW
        msv2.putcol("UVW", np.broadcast_to([1.0, 0.5, 1.5], (nrows, 3)))

        # (UVW2)

        # (DATA)

        # (FLOAT_DATA)

        # (VIDEO_POINT)

        # (LAG_DATA)

        # SIGMA
        msv2.putcol("SIGMA", np.broadcast_to(1.0, (nrows, npols)))

        # (SIGMA_SPECTRUM)

        # WEIGHT
        # msv2.putcol("WEIGHT", np.broadcast_to(1.0, (nrows, npols)))

        # (WEIGHT_SPECTRUM)

        msv2.putcol("FLAG", np.broadcast_to(False, (nrows, nchans, npols)))

        # FLAG_CATEGORY: leave empty

        msv2.putcol("FLAG_ROW", np.broadcast_to(False, (nrows)))

    return outdescr


def gen_subt_ddi(mspath: str, spws_descr: dict, pol_setup_descr: dict):
    """
    Populates DATA_DESCRIPTION
    Q: spws_descr with IDs in keys? IDs in MSv2 are implicit 0...n-1
    """
    import itertools

    # TODO generate spw-pol
    with tables.table(
        mspath + "::DATA_DESCRIPTION", ack=False, readonly=False
    ) as ddi_tbl:
        nrows = len(spws_descr) * len(pol_setup_descr)
        ddi_tbl.addrows(nrows)

        ddis = list(
            itertools.product(list(spws_descr.values()), pol_setup_descr.values())
        )
        spw_ids = list(zip(*ddis))[0]
        pol_ids = list(zip(*ddis))[1]
        ddi_tbl.putcol("SPECTRAL_WINDOW_ID", spw_ids)
        ddi_tbl.putcol("POLARIZATION_ID", pol_ids)
        ddi_tbl.putcol("FLAG_ROW", np.broadcast_to(False, nrows))


def gen_subt_spw(mspath: str, spw_descr: dict, nchans: int, misbehave=False):
    """
    Populates SPECTRAL_WINDOW
    """

    nspws = len(spw_descr)
    with tables.table(
        mspath + "::SPECTRAL_WINDOW", ack=False, readonly=False
    ) as spw_tbl:
        spw_tbl.addrows(nspws)
        # Not in MSv2
        # spw_tbl.putcol("SPECTRAL_WINDOW_ID", list(spw_descr.keys()))
        if misbehave:
            names = ["" for idx in range(nspws)]
        else:
            names = [f"unspecified_test#{idx}" for idx in range(nspws)]
        spw_tbl.putcol("NAME", names)
        # spw_tbl.putcol("REF_FREQUENCY", nspws*[0.9e9])
        # spw_tbl.putcol("TOTAL_BANDWIDTH", nspws*[0.020])

        for spw in range(nspws):
            # nchans = spw_descr[spw]['NUM_CHAN']
            # spw_tbl.addrows(1)  # 1

            # spw_tbl.putcell("NAME", spw, "unspecified_test")
            spw_tbl.putcell("REF_FREQUENCY", spw, 0.9e9)

            spw_tbl.putcol("NUM_CHAN", nchans, startrow=spw, nrow=1)
            widths = np.full(nchans, 20000.0)
            spw_tbl.putcell("CHAN_WIDTH", spw, widths)
            if misbehave:
                kws = spw_tbl.getcolkeywords("CHAN_WIDTH")
                units_kw = "QuantumUnits"
                if units_kw in kws:
                    spw_tbl.removecolkeyword("CHAN_WIDTH", "QuantumUnits")
                # or alternatively
                # spw_tbl.putcol("CHAN_WIDTH", np.full((1, nchans), 20000.0), startrow=spw, nrow=1)
            spw_tbl.putcell("CHAN_FREQ", spw, np.full(nchans, 10e0))
            spw_tbl.putcell("EFFECTIVE_BW", spw, widths)
            spw_tbl.putcell("TOTAL_BANDWIDTH", spw, sum(widths))
            spw_tbl.putcell("RESOLUTION", spw, widths)


def gen_subt_pol_setup(mspath: str, npols, pol_setup_descr: dict):
    """
    populates POLARIZATION
    """

    nsetups = len(pol_setup_descr)

    with tables.table(mspath + "::POLARIZATION", ack=False, readonly=False) as pol_tbl:
        pol_tbl.addrows(npols)
        # pol_tbl.putcol("POLARIZATION_ID", list(spw_descr.keys()))
        pol_tbl.putcol("NUM_CORR", nsetups * [npols])
        pol_tbl.putcol("FLAG_ROW", nsetups * [False])
        corr_types = [9, 11, 13, 15]
        pol_tbl.putcol(
            "CORR_TYPE", np.broadcast_to(corr_types[:npols], (nsetups, npols))
        )
        corr_products = [0, 0, 1, 1]
        pol_tbl.putcol(
            "CORR_PRODUCT",
            np.broadcast_to(corr_products[:npols], (nsetups, nsetups, npols)),
        )


def gen_subt_antenna(mspath: str, ant_descr: dict):
    """
    create ANTENNA
    """

    with tables.table(mspath + "::ANTENNA", ack=False, readonly=False) as ant_tbl:
        nants = len(ant_descr)
        ant_tbl.addrows(nants)
        ant_tbl.putcol("NAME", [f"test_antenna_{idx}" for idx in range(nants)])
        ant_tbl.putcol("STATION", [f"test_station_{idx}" for idx in range(nants)])
        ant_tbl.putcol("DISH_DIAMETER", nants * [12])
        ant_tbl.putcol("FLAG_ROW", nants * [False])
        ant_tbl.putcol("POSITION", np.broadcast_to([0.01, 0.02, 0.03], (nants, 3)))


# field and state very relevant for partitions/sub-MSv2
def gen_subt_field(
    mspath: str, fields_descr: dict, gen_ephem: bool = True, misbehave: bool = False
):
    """
    creates FIELD

    only supports polynomials with 1 coefficient

    Parameters
    ----------
    mspath : str
        path of output MS
    fields_descr : dict
        fields description
    gen_ephem : bool
        whether to generate ephemeris fields (with EPHEM pseudo-sub tables
    misbehave : bool
        if True, some missing SOURCE_IDs will be added

    Returns
    -------

    """

    with tables.table(mspath + "::FIELD", ack=False, readonly=False) as fld_tbl:
        nfields = len(fields_descr)
        fld_tbl.addrows(nfields)
        fld_tbl.putcol("NAME", nfields * ["NGC3031"])

        npoly = 0
        fld_tbl.putcol("NUM_POLY", nfields * [npoly])

        adir = np.deg2rad([30, 45])
        for dir_col in ["DELAY_DIR", "PHASE_DIR", "REFERENCE_DIR"]:
            fld_tbl.putcol(dir_col, np.broadcast_to(adir, (nfields, npoly + 1, 2)))

        fld_tbl.putcol("SOURCE_ID", np.arange(nfields))
        if misbehave:
            fld_tbl.putcell("SOURCE_ID", nfields - 1, nfields + 3)

        if gen_ephem:
            tabdesc_ephemeris_id = {
                "EPHEMERIS_ID": {
                    "valueType": "int",
                    "dataManagerType": "StandardStMan",
                    "dataManagerGroup": "StandardStMan",
                    "option": 0,
                    "maxlen": 0,
                    "ndim": 0,
                    "comment": "comment...",
                }
            }
            fld_tbl.addcols(tabdesc_ephemeris_id)
            fld_tbl.putcol("EPHEMERIS_ID", np.broadcast_to(0, (nfields, 1)))

    if gen_ephem:
        gen_subt_ephem(mspath, misbehave)


def gen_subt_ephem(mspath: str, misbehave=False):
    """
    Creates a phony ephemerides table under the FIELD subtable.
    The usual tables... context or 'default_ms_subtable' not working. Needs
    manual coldesc /not in MSv2/3

    Parameters
    ----------
    mspath : str
        path of output MS
    misbehave : bool
        if True, some metadata (keywords) will be missing: posrefsys, and metadata
        of column 'RA'

    Returns
    -------

    """
    # with open_opt_subtable(mspath, "FIELD/EPHEM0_FIELD.tab") as wtbl:
    ephem0_path = Path(mspath) / "FIELD" / "EPHEM0_FIELDNAME.tab"
    tabdesc = {
        "MJD": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "QuantumUnits": ["s"],
                "MEASINFO": {"type": "epoch", "Ref": "bogus MJD"},
            },
        },
        "RA": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "deg",
            },
        },
        "DEC": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "deg",
            },
        },
        "Rho": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "AU",
            },
        },
        "RadVel": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "AU/d",
            },
        },
        "DiskLong": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "deg",
            },
        },
        "DiskLat": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "deg",
            },
        },
        "SI_lon": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "deg",
            },
        },
        "SI_lat": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "deg",
            },
        },
        "r": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "AU",
            },
        },
        "rdot": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "km/s",
            },
        },
        "phang": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "deg",
            },
        },
    }

    keywords = {
        "VS_CREATE": "2021/01/02/12:33",
        "VS_DATE": "2021/01/02/12:33",
        "VS_VERSION": "0001.0001",
        "MJD0": 58941.4,
        "dMJD": 0.0138889,
        "NAME": "Test_ephem_object",
        "obsloc": "GEOCENTRIC",
        "GeoLong": 0.0,
        "GeoLat": 0.0,
        "GeoDist": 0.0,
        "posrefsys": "ICRF/ICRS",
    }
    with tables.table(
        str(ephem0_path), tabledesc=tabdesc, nrow=1, readonly=False, ack=False
    ) as tbl:

        if misbehave:
            del tabdesc["RA"]["keywords"]
            del keywords["posrefsys"]

        tbl.putcol("MJD", 50000)
        tbl.putcol("RA", 230.334)
        tbl.putcol("DEC", -15.678)
        tbl.putcol("Rho", 0.55)
        tbl.putcol("RadVel", 0.004)
        tbl.putcol("DiskLong", 333.01)
        tbl.putcol("DiskLat", 4.09)
        tbl.putcol("SI_lon", 338.81)
        tbl.putcol("SI_lat", 2.44)
        tbl.putcol("r", 9.1234)
        tbl.putcol("rdot", -1.234)
        tbl.putcol("phang", 5.6789)
        tbl.putkeywords(keywords)


def gen_subt_state(mspath: str, states_descr: dict):
    """
    populates STATE
    """

    # intents in OBS_MODE strings
    with tables.table(mspath + "::STATE", ack=False, readonly=False) as st_tbl:
        nstates = len(states_descr)
        st_tbl.addrows(nstates)
        st_tbl.putcol("SIG", nstates * [True])
        st_tbl.putcol("REF", nstates * [False])
        st_tbl.putcol("CAL", nstates * [0.0])
        st_tbl.putcol("LOAD", nstates * [0.0])
        # TODO: generate subscan ids for potentially multiple scans
        st_tbl.putcol("SUB_SCAN", 1 + np.arange(nstates))
        st_tbl.putcol("OBS_MODE", nstates * ["scan_intent#subscan_intent"])
        st_tbl.putcol("FLAG_ROW", nstates * [False])


# Other - optional subtables


@contextmanager
def open_opt_subtable(
    mspath: str, tbl_name: str
) -> Generator[tables.table, None, None]:
    """
    Opens an (optional) subtable of an MS. This can open tables not included
    in the default_ms definition

    Parameters
    ----------
    mspath : str
        path of output MS
    tbl_name : str
        name of the subtable (WEATHER, etc.). Requires a known optional MS
        subtable name that has a known table description.

    Returns
    -------
    Generator[tables.table, None, None]
        context for an optional subtable created as per MSv2 specs
    """
    subt_desc = tables.complete_ms_desc(tbl_name)
    # table = tables.table(mspath + "/" + tbl_name, tabledesc=subt_desc,
    #                dminfo=makedminfo(subt_desc), ack=False, readonly=False)
    table = default_ms_subtable(tbl_name, mspath + "/" + tbl_name, subt_desc)
    try:
        yield table
    finally:
        table.close()


def gen_subt_source(mspath: str, src_descr: dict, misbehave: bool = False):
    """
    Populate SOURCE subtable, with time dependent source info

    Parameters
    ----------
    mspath : str
        path of output MS
    src_descr : dict
        sources description
    misbehave : bool
        if misbehave, the NUM_LINES and related columns will be missing

    Returns
    -------

    """

    # SOURCE is not included in default_ms
    # with tables.table(mspath + "::SOURCE", ack=False, readonly=False) as src_tbl:
    with open_opt_subtable(mspath, "SOURCE") as src_tbl:
        nsrcs = len(src_descr)
        src_tbl.addrows(nsrcs)
        # if misbehave:
        #    src_tbl.putcol("SOURCE_ID", np.repeat(-1, len(src_descr)))
        src_tbl.putcol("SOURCE_ID", list(src_descr.values()))
        src_tbl.putcol("TIME", np.broadcast_to(0, (nsrcs)))
        src_tbl.putcol("INTERVAL", np.broadcast_to(0, (nsrcs)))
        src_tbl.putcol("SPECTRAL_WINDOW_ID", np.broadcast_to(0, (nsrcs)))

        if not misbehave:
            nlines = 3
            src_tbl.putcol("NUM_LINES", np.repeat(nlines, nsrcs))

        src_tbl.putcol("NAME", np.broadcast_to("test_source_name", (nsrcs)))
        src_tbl.putcol("CALIBRATION_GROUP", np.broadcast_to(0, (nsrcs)))
        src_tbl.putcol("CODE", np.broadcast_to("test_source", (nsrcs)))
        adir = np.deg2rad([29, 34])
        src_tbl.putcol("DIRECTION", np.broadcast_to(adir, (nsrcs, 2)))
        src_tbl.putcol("PROPER_MOTION", np.broadcast_to(adir / 100.0, (nsrcs, 2)))
        # optional cols:
        if not misbehave:
            src_tbl.putcol(
                "TRANSITION",
                np.broadcast_to(
                    ["test_line0", "test_line1", "test_line2"], (nsrcs, nlines)
                ),
            )
            src_tbl.putcol(
                "REST_FREQUENCY", np.broadcast_to([1e9, 2e9, 3e9], (nsrcs, nlines))
            )
            src_tbl.putcol("SYSVEL", np.broadcast_to([1e3, 2e3, 3e3], (nsrcs, nlines)))
        # SOURCE_MODEL is optional and its type is TableRecord!
        src_tbl.removecols(["SOURCE_MODEL"])

    with tables.table(mspath, ack=False, readonly=False) as main:
        main.putkeyword("SOURCE", f"Table: {mspath}/SOURCE")


def gen_subt_pointing(mspath: str):
    """
    Populate POINTING subtable, with antenna-based pointing info

    Very rudimentary.
    """
    # with open_opt_subtable(mspath, "POINTING") as tbl:
    nrows = 1
    with tables.table(mspath + "::POINTING", ack=False, readonly=False) as tbl:
        tbl.addrows(nrows)
        tbl.putcol("ANTENNA_ID", 0)
        tbl.putcol("TIME", 2e9)
        tbl.putcol("INTERVAL", 1e12)
        tbl.putcol("NAME", "test_pointing_name")
        tbl.putcol("NUM_POLY", 0)
        tbl.putcol("TIME_ORIGIN", 1e9)
        adir = np.deg2rad([28.98, 34.03])
        tbl.putcol("DIRECTION", np.broadcast_to(adir, (nrows, 1, 2)))
        tbl.putcol("TARGET", np.broadcast_to(adir + 0.01, (nrows, 1, 2)))
        tbl.putcol("TRACKING", True)


# Other, more secondary subtables


def gen_subt_feed(
    mspath: str,
    feed_descr: dict,
    ant_descr: str,
    spw_descr: str,
    misbehave: bool = False,
):
    """
    Populate FEED subtable, with antenna-based pointing info.

    The table is filled for a single FEED_ID (0), for all the available
    antenna IDs and spectral window IDs. The columns are filled with
    roughtly similar values as seen in example ALMA MSs imported with
    importasdm. BEAM_ID is left as -1 (optional BEAM subtable of MSv2 never
    defined).

    Parameters
    ----------
    mspath : str
        path of output MS
    feed_descr : dict
        feed description, only one feed supported
    ant_descr : str

    spw_descr : str

    misbehave : bool
        if misbehave, the FEED table will be empty.

    Returns
    -------

    """
    if misbehave:
        return

    with tables.table(mspath + "::FEED", ack=False, readonly=False) as tbl:
        nfeeds = len(feed_descr)
        nspws = len(spw_descr)
        nants = len(ant_descr)
        nrows = nants * nspws * nfeeds
        tbl.addrows(nrows)
        tbl.putcol("ANTENNA_ID", np.tile(np.arange(nants), int(nrows / nants)))
        tbl.putcol("FEED_ID", np.repeat(0, (nrows)))
        tbl.putcol(
            "SPECTRAL_WINDOW_ID", np.repeat(list(spw_descr.values()), (nrows / nspws))
        )
        tbl.putcol("TIME", np.broadcast_to(0, (nrows)))
        tbl.putcol("INTERVAL", np.broadcast_to(5e9, (nrows)))
        nrecep = 2
        tbl.putcol("NUM_RECEPTORS", np.broadcast_to(nrecep, (nrows)))
        tbl.putcol("BEAM_ID", np.broadcast_to(-1, (nrows)))
        boff = np.deg2rad([0.1, 0.3])
        tbl.putcol("BEAM_OFFSET", np.broadcast_to(boff, (nrows, 2, nrecep)))
        pol_types = ["test_X, test_Y"]
        len_pol_types = 2
        tbl.putcol(
            "POLARIZATION_TYPE", np.broadcast_to(pol_types, (nrows, len_pol_types))
        )
        tbl.putcol("POL_RESPONSE", np.broadcast_to(0.0, (nrows, 2, nrecep)))
        tbl.putcol("POSITION", np.broadcast_to([0.0, 0.0, 0.0], (nrows, 3)))
        tbl.putcol("RECEPTOR_ANGLE", np.broadcast_to([1.51, 0.33], (nrows, nrecep)))


def gen_subt_observation(mspath: str, obs_descr: dict):
    """
    Populate the OBSERVATION table, with one row per observation id/number
    (MSv2: the id is implicitly the row number).
    The string columns are filled with values loosely based on example ALMA
    MSs.

    Parameters
    ----------
    mspath : str
        path of output MS
    obs_descr : dict
        obs description, with IDs, only the len is considered

    Returns
    -------

    """

    nobs = len(obs_descr)
    with tables.table(mspath + "::OBSERVATION", ack=False, readonly=False) as tbl:
        tbl.addrows(nobs)
        # no keys, all data cols:
        # "test_telescope" would produce errors for example in listobs:
        # Exception: Telescope test_telescope is not recognized by CASA.
        # (casa6core::MPosition casa6core::MSMetaData::getObservatoryPosition())
        tbl.putcol("TELESCOPE_NAME", np.repeat("ALMA", nobs))
        tbl.putcol("TIME_RANGE", np.broadcast_to([[4.87021e9, 4.87022e9]], (nobs, 2)))
        tbl.putcol("OBSERVER", np.repeat("Dr. test_observer", nobs))
        tbl.putcol(
            "LOG", np.broadcast_to(["test_obs_log_1", "test_obs_log_2"], (nobs, 2))
        )
        tbl.putcol("SCHEDULE_TYPE", np.repeat("test_schedule_type", nobs))
        sched = np.array(
            [
                [
                    f"SchedulingBlock uid://A002/X1ftest/Xde{idx}",
                    f"ExecBlock uid://A002/X1abtest/X123{idx}",
                ]
                for idx in range(nobs)
            ]
        )
        tbl.putcol("SCHEDULE", sched)
        proj = np.array([f"uid://A002/X1ftest/X4ec{idx}" for idx in range(nobs)])
        tbl.putcol("PROJECT", proj)
        tbl.putcol("RELEASE_DATE", np.repeat(0, nobs))


def gen_subt_processor(mspath: str, proc_descr: dict):
    """
    Populate the PROCESSOR table, with one row per processor id/number
    (MSv2: the id is implicitly the row number). The TYPE_ID is left to -1.
    All values of TYPE are CORRELATOR.
    MODE_ID are also left to -1 (no ..._MODE subtable).
    The string columns are filled with values loosely based on example ALMA
    MSs.

    Parameters
    ----------
    mspath : str
        path of output MS
    proc_descr : dict
        processors description, with IDs, only the len is considered

    Returns
    -------

    """

    nproc = len(proc_descr)
    with tables.table(mspath + "::PROCESSOR", ack=False, readonly=False) as tbl:
        tbl.addrows(nproc)
        # no keys, all data cols:
        # There could also be RADIOMETER, SPECTROMETER, PULSAR-TIMER, etc. - not nor now
        tbl.putcol("TYPE", np.repeat("CORRELATOR", nproc))
        tbl.putcol("SUB_TYPE", np.repeat("test_CORRELATOR_MODE", nproc))
        tbl.putcol("TYPE_ID", np.repeat(-1, nproc))
        tbl.putcol("MODE_ID", np.repeat(-1, nproc))
        tbl.putcol("FLAG_ROW", np.repeat(False, nproc))


def gen_subt_flag_cmd(mspath: str):
    """
    Leaves the FLAG_CMD subtable empty for now, and checks that there are no
    rows.
    """
    with tables.table(mspath + "::FLAG_CMD", ack=False, readonly=False) as tbl:
        assert tbl.nrows() == 0


def gen_subt_history(mspath: str, obs_descr: dict):
    """
    Populate the HISTORY table, with only one row per observation
    The string columns are filled with values loosely based on example ALMA MSs.
    OBJECT_ID is left all to -1.

    Parameters
    ----------
    mspath : str
        path of output MS
    obs_descr : dict
        obs description, with IDs, only the len is considered

    Returns
    -------

    """

    nobs = len(obs_descr)
    with tables.table(mspath + "::HISTORY", ack=False, readonly=False) as tbl:
        tbl.addrows(nobs)
        # keys
        tbl.putcol("TIME", np.repeat(0, nobs))
        tbl.putcol("OBSERVATION_ID", np.arange(nobs))
        # data
        tbl.putcol("MESSAGE", np.repeat("made with test ms generator", nobs))
        tbl.putcol("PRIORITY", np.repeat("INFO", nobs))
        tbl.putcol("ORIGIN", np.repeat("ms_maker", nobs))
        tbl.putcol("OBJECT_ID", np.repeat(0, nobs))
        tbl.putcol("APPLICATION", np.repeat("xradio", nobs))
        tbl.putcol("CLI_COMMAND", np.broadcast_to("gen_subt_history", (nobs, 1)))
        tbl.putcol("APP_PARAMS", np.broadcast_to("", (nobs, 1)))


def gen_subt_syscal(mspath: str, ant_descr: dict):
    """
    Creates a SYSCAL subtable and populates it with a (very incomplete) row
    This is just to enable minimal coverage of some SYSCAL handling code in
    the casacore tables read/write functions.
    """
    ncal = len(ant_descr)
    subt_name = "SYSCAL"
    with open_opt_subtable(mspath, subt_name) as sctbl:
        sctbl.addrows(ncal)
        sctbl.putcol("ANTENNA_ID", np.arange(0, ncal))
        sctbl.putcol("FEED_ID", np.repeat(0, ncal))
        sctbl.putcol("SPECTRAL_WINDOW_ID", np.repeat(0, ncal))
        sctbl.putcol("TIME", np.repeat(1e10, ncal))
        sctbl.putcol("INTERVAL", np.repeat(1e12, ncal))
        # all data/flags columns in the SYSCAL table are optional!
        # sctbl.putcol("PHASE_DIFF", np.repeat(0.3, ncal))
        sctbl.putcol("PHASE_DIFF", np.repeat(0.3, ncal))
        # sctbl.putcol("TCAL", np.broadcast_to(50.3, (ncal, 2)))
        sctbl.putcol("TCAL_SPECTRUM", np.broadcast_to(50.3, (ncal, 10, 2)))
        # TRX, etc.

    with tables.table(mspath, ack=False, readonly=False) as main:
        main.putkeyword(subt_name, f"Table: {mspath}/{subt_name}")


def gen_subt_weather(mspath: str, misbehave: bool = False):
    """
    Creates a WEATHER subtable and populates it with a (very incomplete) row
    simply with a very high TIME value, as seen in some corner cases of
    test MSs.
    This is just to enable minimal coverage of some WEATHER handling code in
    the casacore tables read/write functions.
    """
    subt_name = "WEATHER"
    tabdesc_ns_wx_station = {
        "NS_WX_STATION_ID": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "NS_WX_STATION_POSITION": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "ndim": 1,
            "comment": "comment...",
            "keywords": {},
        },
    }
    nrows = 5
    with open_opt_subtable(mspath, subt_name) as wtbl:
        wtbl.addrows(nrows)
        wtbl.putcol("ANTENNA_ID", np.arange(0, nrows))
        wtbl.putcol("TIME", np.arange(0, nrows) * 100 + 1e12)
        wtbl.putcol("INTERVAL", np.repeat(1e12, nrows))
        # all data/flags columns in the WEATHER table are optional!
        # But note they are always added in the python-casacore defaults
        wtbl.putcol("H2O", np.repeat(0.03, nrows))
        if not misbehave:
            wtbl.addcols(tabdesc_ns_wx_station)
            wtbl.putcol("ANTENNA_ID", np.repeat(-1, nrows))
            wtbl.putcol("NS_WX_STATION_ID", np.arange(0, nrows))
            wtbl.putcol(
                "NS_WX_STATION_POSITION", np.broadcast_to([0.1, 0.2, 0.3], (nrows, 3))
            )

    with tables.table(mspath, ack=False, readonly=False) as main:
        main.putkeyword(subt_name, f"Table: {mspath}/{subt_name}")


def gen_subt_asdm_receiver(mspath: str):
    """
    Produces an over-simple table, with only one row, for basic coverage of ASDM_* subtables handling
    code.
    """

    subt_name = "ASDM_RECEIVER"
    rec_path = Path(mspath) / subt_name
    tabdesc = {
        "receiverId": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "spectralWindowId": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "timeInterval": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
    }

    with tables.table(
        str(rec_path), tabledesc=tabdesc, nrow=1, readonly=False, ack=False
    ) as tbl:
        tbl.putcol("receiverId", 0)
        tbl.putcol("spectralWindowId", 0)
        tbl.putcol("timeInterval", 0)

    with tables.table(mspath, ack=False, readonly=False) as main:
        main.putkeyword(subt_name, f"Table: {mspath}/{subt_name}")


def gen_subt_asdm_station(mspath: str):
    """
    Produces an over-simple table, with only a few rows, for basic coverage of ASDM_STATION subtable
    handling (relevant when loading WEATHER)
    """

    subt_name = "ASDM_STATION"
    rec_path = Path(mspath) / subt_name
    tabdesc = {
        "stationId": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "name": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "position": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "ndim": 1,
            "comment": "comment...",
            "keywords": {},
        },
        "type": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "time": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
    }

    nrows = 5
    with tables.table(
        str(rec_path), tabledesc=tabdesc, nrow=nrows, readonly=False, ack=False
    ) as tbl:
        tbl.putcol("stationId", np.arange(0, nrows))
        tbl.putcol("position", np.broadcast_to([10, 20, 30], (nrows, 3)))
        tbl.putcol("name", [f"test_station{idx}" for idx in np.arange(0, nrows)])
        tbl.putcol(
            "type", np.repeat("WEATHER_STATION", nrows)
        )  # ANTENNA_PAD / MAINTENANCE_PAD
        tbl.putcol("time", np.repeat(1e9, nrows))

    with tables.table(mspath, ack=False, readonly=False) as main:
        main.putkeyword(subt_name, f"Table: {mspath}/{subt_name}")


def gen_subt_asdm_execblock(mspath: str):
    """
    Produces a basic ASDM/EXECBLOCK table, for basic coverage of code that handles the ASDM_*
    subtables.
    For now it simply creates a table with one row
    """

    subt_name = "ASDM_EXECBLOCK"
    rec_path = Path(mspath) / subt_name
    tabdesc = {
        "execBlockIDId": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "execBlockNumId": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "execBlockUID": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "sessionReference": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "observingScript": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "observingScriptUID": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "observingLog": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "ndim": 1,
            "_c_order": True,
            "keywords": {},
        },
    }

    nrows = 1
    with tables.table(
        str(rec_path), tabledesc=tabdesc, nrow=nrows, readonly=False, ack=False
    ) as tbl:
        tbl.putcol("execBlockIDId", "1")
        tbl.putcol("execBlockNumId", "3")
        tbl.putcol("execBlockUID", "uid://A001/X1abtest/X123")
        tbl.putcol("sessionReference", "test_session_ref")
        tbl.putcol("observingScript", "test script")
        tbl.putcol("observingScriptUID", "uid://A003/X1abtest/X987")
        tbl.putcol("observingLog", np.broadcast_to("test log line 0", (nrows, 1)))
        # tbl.putcol("observingLog", np.broadcast_to(np.array(["test log line 0", "test log line 1"], dtype="str"), (nrows, 2)))

    with tables.table(mspath, ack=False, readonly=False) as main:
        main.putkeyword(subt_name, f"Table: {mspath}/{subt_name}")


def gen_subt_gain_curve(mspath: str, ant_descr: dict):
    """
    Produces a basic GAIN_CURVE (sub)table, following casacore note #265, for basic coverage of
    VLBI subtables handling.
    """

    subt_name = "GAIN_CURVE"
    rec_path = Path(mspath) / subt_name
    tabdesc = {
        "ANTENNA_ID": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "FEED_ID": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "SPECTRAL_WINDOW_ID": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "TIME": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "s",
            },
        },
        "INTERVAL": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "s",
            },
        },
        "TYPE": {
            "valueType": "string",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "NUM_POLY": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "GAIN": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "ndim": 2,
            # "shape"
            "comment": "comment...",
            "keywords": {},
        },
        "SENSITIVITY": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "ndim": 1,
            # "shape"
            "comment": "comment...",
            "keywords": {
                "UNIT": "K/Jy",
            },
        },
    }

    nants = len(ant_descr)
    nreceptors = 2
    num_poly = 2
    with tables.table(
        str(rec_path), tabledesc=tabdesc, nrow=nants, readonly=False, ack=False
    ) as tbl:
        tbl.putcol("ANTENNA_ID", np.arange(0, nants))
        tbl.putcol("FEED_ID", np.repeat(0, nants))
        tbl.putcol("SPECTRAL_WINDOW_ID", np.repeat(0, nants))
        tbl.putcol("TIME", np.repeat(1e12, nants))
        tbl.putcol("INTERVAL", np.repeat(2, nants))
        tbl.putcol("TYPE", np.repeat("(”POWER(EL)", nants))
        tbl.putcol("NUM_POLY", np.repeat(num_poly, nants))
        tbl.putcol("GAIN", np.broadcast_to(0.85, (nants, nreceptors, num_poly)))
        tbl.putcol("SENSITIVITY", np.broadcast_to(0.95, (nants, nreceptors)))

    with tables.table(mspath, ack=False, readonly=False) as main:
        main.putkeyword(subt_name, f"Table: {mspath}/{subt_name}")


def gen_subt_phase_cal(mspath: str, ant_descr: dict):
    """
    Produces a basic PHASE_CAL (sub)table, following casacore note #265, for basic coverage of
    VLBI subtables handling.
    Note some differences in example test datasets like VLBA_TL016B_split.ms dimensions with respect
    to the casacore note tables.
    """

    subt_name = "PHASE_CAL"
    rec_path = Path(mspath) / subt_name
    tabdesc = {
        "ANTENNA_ID": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "FEED_ID": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "SPECTRAL_WINDOW_ID": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "TIME": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "s",
            },
        },
        "INTERVAL": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {
                "UNIT": "s",
            },
        },
        "NUM_TONES": {
            "valueType": "int",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "comment": "comment...",
            "keywords": {},
        },
        "TONE_FREQUENCY": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "ndim": 2,
            # "shape":
            "comment": "comment...",
            "keywords": {
                "QuantumUnits": ["Hz"],
                "MEASINFO": {"type": "frequency", "Ref": "bogus ref frame"},
            },
        },
        "PHASE_CAL": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            "ndim": 2,
            # "shape"
            "comment": "comment...",
            "keywords": {},
        },
        "CABLE_CAL": {
            "valueType": "double",
            "dataManagerType": "StandardStMan",
            "dataManagerGroup": "StandardStMan",
            "option": 0,
            "maxlen": 0,
            # "ndim": 1,
            # "shape"
            "comment": "comment...",
            "keywords": {
                "QuantumUnits": "s",
            },
        },
    }

    nants = len(ant_descr)
    nreceptors = 2
    ntones = 3
    _ntimes = 1
    with tables.table(
        str(rec_path), tabledesc=tabdesc, nrow=nants, readonly=False, ack=False
    ) as tbl:
        tbl.putcol("ANTENNA_ID", np.arange(0, nants))
        tbl.putcol("FEED_ID", np.repeat(0, nants))
        tbl.putcol("SPECTRAL_WINDOW_ID", np.repeat(0, nants))
        tbl.putcol("TIME", np.repeat(1e12, nants))
        tbl.putcol("INTERVAL", np.repeat(2, nants))
        tbl.putcol("NUM_TONES", np.repeat(ntones, nants))
        tbl.putcol(
            "TONE_FREQUENCY", np.broadcast_to(1.234e9, (nants, ntones, nreceptors))
        )
        tbl.putcol("PHASE_CAL", np.broadcast_to(0.92, (nants, ntones, nreceptors)))
        tbl.putcol("CABLE_CAL", np.repeat(0.93, (nants)))

    with tables.table(mspath, ack=False, readonly=False) as main:
        main.putkeyword(subt_name, f"Table: {mspath}/{subt_name}")


# Functions from io.py that depend on casacore


def build_processing_set_from_msv2(
    in_file: str | Path,
    out_file: str | Path,
    *,
    partition_scheme: Optional[Iterable[dict]] = None,
    persistence_mode: str = "w",
    parallel_mode: str = "partition",
    **convert_kwargs: Any,
) -> Path:
    """
    Convert an MSv2 dataset into a processing set using the production converter.
    """
    from xradio.measurement_set import convert_msv2_to_processing_set

    convert_msv2_to_processing_set(
        in_file=str(in_file),
        out_file=str(out_file),
        partition_scheme=list(partition_scheme or []),
        persistence_mode=persistence_mode,
        parallel_mode=parallel_mode,
        **convert_kwargs,
    )
    return Path(out_file)


def build_msv4_partition(
    ms_path: str | Path,
    out_root: str | Path,
    *,
    msv4_id: str = "msv4id",
    partition_kwargs: Optional[Dict[str, Any]] = None,
    use_table_iter: bool = False,
    persistence_mode: str = "w",
) -> Path:
    """
    Convert a MeasurementSet v2 partition into an MSv4 Zarr tree.
    """
    from xradio.measurement_set._utils._msv2.conversion import (
        convert_and_write_partition,
    )

    partition_kwargs = partition_kwargs or {"DATA_DESC_ID": [0]}
    convert_and_write_partition(
        str(ms_path),
        str(out_root),
        msv4_id,
        partition_kwargs,
        use_table_iter=use_table_iter,
        persistence_mode=persistence_mode,
    )
    return Path(out_root) / f"{Path(ms_path).stem}_{msv4_id}"


def build_minimal_msv4_xdt(
    ms_path: str | Path,
    *,
    out_root: str | Path | None = None,
    msv4_id: str = "msv4id",
    partition_kwargs: Optional[Dict[str, Any]] = None,
    use_table_iter: bool = False,
    persistence_mode: str = "w",
) -> Path:
    """
    Convenience wrapper that selects reasonable defaults for the minimal MSv4 conversion.
    """

    if out_root is None:
        out_root = Path(f"{Path(ms_path).stem}_processing_set.zarr")
    return build_msv4_partition(
        ms_path,
        out_root,
        msv4_id=msv4_id,
        partition_kwargs=partition_kwargs,
        use_table_iter=use_table_iter,
        persistence_mode=persistence_mode,
    )



import itertools
import time
import os
import pandas as pd

import numpy as np

try:
    from casacore import tables
except ImportError:
    import xradio._utils._casacore.casacore_from_casatools as tables

from xradio.measurement_set._utils._msv2._tables.read import table_exists
from xradio._utils.logging import xradio_logger


def enumerated_product(*args):
    yield from zip(
        itertools.product(*(range(len(x)) for x in args)), itertools.product(*args)
    )


import pickle, gzip


def create_partitions(in_file: str, partition_scheme: list) -> list[dict]:
    """Create a list of dictionaries with the partition information.

    Parameters
    ----------
    in_file: str
        Input MSv2 file path.
    partition_scheme:  list
        A MS v4 can only contain a single data description (spectral window and polarization setup), and observation mode. Consequently, the MS v2 is partitioned when converting to MS v4.
        In addition to data description and polarization setup a finer partitioning is possible by specifying a list of partitioning keys. Any combination of the following keys are possible:
        "FIELD_ID", "SCAN_NUMBER", "STATE_ID", "SOURCE_ID", "SUB_SCAN_NUMBER", "ANTENNA1".
        For mosaics where the phase center is rapidly changing (such as VLA on the fly mosaics)  partition_scheme should be set to an empty list []. By default, ["FIELD_ID"].
    Returns
    -------
    list
        list of dictionaries with the partition information.
    """

    ### Test new implementation without
    # Always start with these (if available); then extend with user scheme.
    partition_scheme = [
        "DATA_DESC_ID",
        "OBS_MODE",
        "OBSERVATION_ID",
        "EPHEMERIS_ID",
    ] + list(partition_scheme)

    # partition_scheme = ["DATA_DESC_ID", "OBS_MODE"] + list(
    #     partition_scheme
    # )

    t0 = time.time()
    # --------- Load base columns from MAIN table ----------
    main_tb = tables.table(
        in_file, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )

    # Build minimal DF once. Pull only columns we may need.
    # Add columns here if you expect to aggregate them per-partition.
    base_cols = {
        "DATA_DESC_ID": main_tb.getcol("DATA_DESC_ID"),
        "FIELD_ID": main_tb.getcol("FIELD_ID"),
        "SCAN_NUMBER": main_tb.getcol("SCAN_NUMBER"),
        "STATE_ID": main_tb.getcol("STATE_ID"),
        "OBSERVATION_ID": main_tb.getcol("OBSERVATION_ID"),
        "ANTENNA1": main_tb.getcol("ANTENNA1"),
    }
    par_df = pd.DataFrame(base_cols).drop_duplicates()
    xradio_logger().debug(
        f"Loaded MAIN columns in {time.time() - t0:.2f}s "
        f"({len(par_df):,} unique MAIN rows)"
    )

    # --------- Optional SOURCE/STATE derived columns ----------
    # SOURCE_ID (via FIELD table)
    t1 = time.time()
    source_id_added = False
    field_tb = tables.table(
        os.path.join(in_file, "FIELD"),
        readonly=True,
        lockoptions={"option": "usernoread"},
        ack=False,
    )
    if table_exists(os.path.join(in_file, "SOURCE")):
        source_tb = tables.table(
            os.path.join(in_file, "SOURCE"),
            readonly=True,
            lockoptions={"option": "usernoread"},
            ack=False,
        )
        if source_tb.nrows() != 0:
            # Map SOURCE_ID via FIELD_ID
            field_source = np.asarray(field_tb.getcol("SOURCE_ID"))
            par_df["SOURCE_ID"] = field_source[par_df["FIELD_ID"]]
            source_id_added = True
    xradio_logger().debug(
        f"SOURCE processing in {time.time() - t1:.2f}s "
        f"(added SOURCE_ID={source_id_added})"
    )

    if "EPHEMERIS_ID" in field_tb.colnames():
        ephemeris_id_added = False
        if field_tb.nrows() != 0:
            # Map EPHEMERIS_ID via FIELD_ID
            field_ephemeris = np.asarray(field_tb.getcol("EPHEMERIS_ID"))
            par_df["EPHEMERIS_ID"] = field_ephemeris[par_df["FIELD_ID"]]
            ephemeris_id_added = True
        xradio_logger().debug(
            f"EPHEMERIS processing in {time.time() - t1:.2f}s "
            f"(added EPHEMERIS_ID={ephemeris_id_added})"
        )

    # OBS_MODE & SUB_SCAN_NUMBER (via STATE table)
    t2 = time.time()
    obs_mode_added = False
    sub_scan_added = False
    if table_exists(os.path.join(in_file, "STATE")):
        state_tb = tables.table(
            os.path.join(in_file, "STATE"),
            readonly=True,
            lockoptions={"option": "usernoread"},
            ack=False,
        )
        if state_tb.nrows() != 0:
            state_obs_mode = np.asarray(state_tb.getcol("OBS_MODE"))
            state_sub_scan = np.asarray(state_tb.getcol("SUB_SCAN"))
            # Index by STATE_ID into STATE columns
            par_df["OBS_MODE"] = state_obs_mode[par_df["STATE_ID"]]
            par_df["SUB_SCAN_NUMBER"] = state_sub_scan[par_df["STATE_ID"]]
            obs_mode_added = True
            sub_scan_added = True
        else:
            # If STATE empty, drop STATE_ID (it cannot partition anything)
            if "STATE_ID" in par_df.columns:
                par_df.drop(columns=["STATE_ID"], inplace=True)

            if "SUB_SCAN_NUMBER" in par_df.columns:
                par_df.drop(columns=["SUB_SCAN_NUMBER"], inplace=True)

    xradio_logger().debug(
        f"STATE processing in {time.time() - t2:.2f}s "
        f"(OBS_MODE={obs_mode_added}, SUB_SCAN_NUMBER={sub_scan_added})"
    )

    # --------- Decide which partition keys are actually available ----------
    t3 = time.time()
    partition_scheme_updated = [k for k in partition_scheme if k in par_df.columns]
    xradio_logger().info(f"Updated partition scheme used: {partition_scheme_updated}")

    # If none of the requested keys exist, there is a single partition of "everything"
    if not partition_scheme_updated:
        partition_scheme_updated = []

    # These are the axes we report per partition (present => aggregate unique values)
    partition_axis_names = [
        "DATA_DESC_ID",
        "OBSERVATION_ID",
        "FIELD_ID",
        "SCAN_NUMBER",
        "STATE_ID",
        "SOURCE_ID",
        "OBS_MODE",
        "SUB_SCAN_NUMBER",
        "EPHEMERIS_ID",
    ]
    # Only include ANTENNA1 if user asked for it (keeps output size down)
    if "ANTENNA1" in partition_scheme:
        partition_axis_names.append("ANTENNA1")

    # --------- Group only by realized partitions (no Cartesian product!) ----------
    # observed=True speeds up if categorical; here it’s harmless. sort=False keeps source order.
    if partition_scheme_updated:
        grp = par_df.groupby(partition_scheme_updated, sort=False, observed=False)
        groups_iter = grp
    else:
        # Single group: everything
        groups_iter = [(None, par_df)]

    partitions = []
    # Fast aggregation: use NumPy for uniques to avoid pandas overhead in the tight loop.
    for _, gdf in groups_iter:
        part = {}
        for name in partition_axis_names:
            if name in gdf.columns:
                # Return Python lists to match your prior structure (can be np.ndarray if preferred)
                part[name] = np.unique(gdf[name].to_numpy()).tolist()
            else:
                part[name] = [None]
        partitions.append(part)

    xradio_logger().debug(
        f"Partition build in {time.time() - t3:.2f}s; total {len(partitions):,} partitions"
    )
    xradio_logger().debug(f"Total create_partitions time: {time.time() - t0:.2f}s")

    # # with gzip.open("partition_original_small.pkl.gz", "wb") as f:
    # #     pickle.dump(partitions, f, protocol=pickle.HIGHEST_PROTOCOL)

    # #partitions[1]["DATA_DESC_ID"] = [999]  # make a change to test comparison
    # #org_partitions = load_dict_list("partition_original_small.pkl.gz")
    # org_partitions = load_dict_list("partition_original.pkl.gz")

    return partitions


from typing import Any, List, Dict


def save_dict_list(filename: str, data: List[Dict[str, Any]]) -> None:
    """
    Save a list of dictionaries containing NumPy arrays (or other objects)
    to a compressed pickle file.
    """
    with gzip.open(filename, "wb") as f:
        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)


def load_dict_list(filename: str) -> List[Dict[str, Any]]:
    """
    Load a list of dictionaries containing NumPy arrays (or other objects)
    from a compressed pickle file.
    """
    with gzip.open(filename, "rb") as f:
        return pickle.load(f)


def dict_list_equal(a: List[Dict[str, Any]], b: List[Dict[str, Any]]) -> bool:
    """
    Compare two lists of dictionaries to ensure they are exactly the same.
    NumPy arrays are compared with array_equal, other objects with ==.
    """
    if len(a) != len(b):
        return False

    for d1, d2 in zip(a, b):
        if d1.keys() != d2.keys():
            return False
        for k in d1:
            v1, v2 = d1[k], d2[k]
            if isinstance(v1, np.ndarray) and isinstance(v2, np.ndarray):
                if not np.array_equal(v1, v2):
                    return False
            else:
                if v1 != v2:
                    return False
    return True


from typing import Iterable, Mapping, Tuple, List, Dict, Any, Set
import numpy as np


def _to_python_scalar(x: Any) -> Any:
    """Convert NumPy scalars to Python scalars; leave others unchanged."""
    if isinstance(x, np.generic):
        return x.item()
    return x


def _to_hashable_value_list(v: Any) -> Tuple[Any, ...]:
    """
    Normalize a dict value (often list/np.ndarray) into a sorted, hashable tuple.
    - Accepts list/tuple/np.ndarray/scalars/None.
    - Treats None as a value.
    - Sorts with a stable key that stringifies items to avoid dtype hiccups.
    """
    if isinstance(v, np.ndarray):
        v = v.tolist()
    if v is None or isinstance(v, (str, bytes)):
        # Treat a bare scalar as a single-element collection for consistency.
        v = [v]
    elif not isinstance(v, (list, tuple)):
        v = [v]

    py_vals = [_to_python_scalar(x) for x in v]
    # Sort by (type name, repr) to keep mixed types stable if present
    return tuple(sorted(py_vals, key=lambda x: (type(x).__name__, repr(x))))


def _canon_partition(
    d: Mapping[str, Any], ignore_keys: Iterable[str] = ()
) -> Tuple[Tuple[str, Tuple[Any, ...]], ...]:
    """
    Canonicalize a partition dict into a hashable, order-insensitive representation.
    - Drops keys in ignore_keys.
    - Converts each value collection to a sorted tuple.
    - Sorts keys.
    """
    ign: Set[str] = set(ignore_keys)
    items = []
    for k, v in d.items():
        if k in ign:
            continue
        items.append((k, _to_hashable_value_list(v)))
    items.sort(key=lambda kv: kv[0])
    return tuple(items)


def compare_partitions_subset(
    new_partitions: List[Dict[str, Any]],
    original_partitions: List[Dict[str, Any]],
    ignore_keys: Iterable[str] = (),
) -> Tuple[bool, List[Dict[str, Any]]]:
    """
    Check that every partition in `new_partitions` also appears in `original_partitions`,
    ignoring ordering (of partitions and of values within each key).

    Parameters
    ----------
    new_partitions : list of dict
        Partitions produced by the optimized/new code.
    original_partitions : list of dict
        Partitions produced by the original code (the reference).
    ignore_keys : iterable of str, optional
        Keys to ignore when comparing partitions (e.g., timestamps or debug fields).

    Returns
    -------
    (ok, missing)
        ok : bool
            True if every new partition is found in the original set.
        missing : list of dict
            The list of partitions (from `new_partitions`) that were NOT found in `original_partitions`,
            useful for debugging diffs.
    """
    orig_set = {_canon_partition(p, ignore_keys) for p in original_partitions}
    missing = []
    for p in new_partitions:
        cp = _canon_partition(p, ignore_keys)
        if cp not in orig_set:
            missing.append(p)
    return (len(missing) == 0, missing)



from types import ModuleType
import logging

try:
    import toolviper.utils.logger as _logger  # noqa: F841
except ImportError:
    _logger = logging


def xradio_logger() -> ModuleType:
    """Returns the toolviper logging module if available,
    otherwise the standard Python logging module"""
    return _logger



subt_rename_ids = {
    "ANTENNA": {"row": "antenna_id", "dim_1": "xyz"},
    "FEED": {"dim_1": "xyz", "dim_2": "receptor", "dim_3": "receptor2"},
    "FIELD": {"row": "field_id", "dim_1": "poly_id", "dim_2": "ra/dec"},
    "FREQ_OFFSET": {"antenna1": "antenna1_id", "antenna2": "antenna2_id"},
    "OBSERVATION": {"row": "observation_id", "dim_1": "start/end"},
    "POINTING": {"dim_1": "n_polynomial", "dim_3": "dir"},
    "POLARIZATION": {"row": "pol_setup_id", "dim_2": "product_id"},
    "PROCESSOR": {"row": "processor_id"},
    "SPECTRAL_WINDOW": {"row": "spectral_window_id", "dim_1": "chan"},
    "SOURCE": {"dim_1": "ra/dec", "dim_2": "line"},
    "STATE": {"row": "state_id"},
    "SYSCAL": {"dim_1": "frequency", "dim_2": "receptor"},
    # Would make sense for non-std "WS_NX_STATION_POSITION"
    "WEATHER": {"dim_1": "xyz"},
}



from xradio._utils.zarr.common import _get_file_system_and_items
import xarray as xr


def open_processing_set(
    ps_store: str,
    scan_intents: list | None = None,
) -> xr.DataTree:
    """Creates a lazy representation of a Processing Set (only meta-data is loaded into memory).

    Parameters
    ----------
    ps_store : str
        String of the path and name of the processing set. For example '/users/user_1/uid___A002_Xf07bba_Xbe5c_target.lsrk.vis.zarr'.
    scan_intents : str | None, optional
        A list of scan_intents to be opened for example ['OBSERVE_TARGET#ON_SOURCE']. The scan_intents in a processing_set_xdt can be seen by calling processing_set_xdt.ps.summary().
        By default None, which will include all scan_intents.

    Returns
    -------
    xarray.DataTree
        Lazy representation of processing set (the data arrays of the datasets are
        represented by Dask.arrays).
    """

    file_system, ms_store_list = _get_file_system_and_items(ps_store)
    import s3fs

    if isinstance(file_system, s3fs.core.S3FileSystem):
        mapping = s3fs.S3Map(root=ps_store, s3=file_system, check=False)
        ps_xdt = xr.open_datatree(
            mapping, engine="zarr", chunks={}, chunked_array_type="dask"
        )
    else:
        ps_xdt = xr.open_datatree(
            ps_store, engine="zarr", chunks={}, chunked_array_type="dask"
        )

    # Future work is to add ASDM backend

    if scan_intents is None:
        return ps_xdt
    else:
        return ps_xdt.xr_ps.query(scan_intents=scan_intents)


# def open_processing_set(
#     ps_store: str,
#     intents: list = None,
# ): #-> ProcessingSet:
#     """Creates a lazy representation of a Processing Set (only meta-data is loaded into memory).

#     Parameters
#     ----------
#     ps_store : str
#         String of the path and name of the processing set. For example '/users/user_1/uid___A002_Xf07bba_Xbe5c_target.lsrk.vis.zarr'.
#     intents : list, optional
#         A list of intents to be open for example ['OBSERVE_TARGET#ON_SOURCE']. The intents in a processing set can be seen by calling processing_set.summary().
#         By default None, which will open all intents.

#     Returns
#     -------
#     processing_set
#         Lazy representation of processing set (data is represented by Dask.arrays).
#     """
#     from xradio.measurement_set import MeasurementSetXds

#     file_system, ms_store_list = _get_file_system_and_items(ps_store)
#     from xradio.measurement_set import ProcessingSet
#     ps = ProcessingSet()
#     data_group = "base"
#     for ms_name in ms_store_list:
#         # try:
#         ms_store = os.path.join(ps_store, ms_name)
#         correlated_store = os.path.join(ms_store, "correlated_xds")

#         xds = _open_dataset(correlated_store, file_system)
#         data_groups = xds.attrs["data_groups"]

#         if (intents is None) or (
#             bool(set(xds.attrs["partition_info"]["intents"]).intersection(intents))
#         ):
#             sub_xds_dict, field_and_source_xds_dict = _open_sub_xds(
#                 ms_store, file_system=file_system, data_groups=data_groups
#             )

#             xds.attrs = {
#                 **xds.attrs,
#                 **sub_xds_dict,
#             }

#             for data_group_name, data_group_vals in data_groups.items():
#                 xds[data_group_vals["correlated_data"]].attrs[
#                     "field_and_source_xds"
#                 ] = field_and_source_xds_dict[data_group_name]

#             ps[ms_name] = MeasurementSetXds(xds)
#         # except Exception as e:
#         #     logger.warning(f"Could not open {ms_name} due to {e}")
#         #     continue

#     return ps


# def _open_sub_xds(ms_store, file_system, data_groups, load=False):
#     sub_xds_dict = {}
#     field_and_source_xds_dict = {}

#     if isinstance(file_system, s3fs.core.S3FileSystem):
#         file_names = [
#             bd.split(sep="/")[-1] for bd in file_system.listdir(ms_store, detail=False)
#         ]
#     else:
#         file_names = file_system.listdir(ms_store)
#     file_names = [item for item in file_names if not item.startswith(".")]

#     file_names.remove("correlated_xds")

#     field_dict = {"field_and_source_xds_" + key: key for key in data_groups.keys()}

#     # field_and_source_xds_name_start = "FIELD"
#     for n in file_names:
#         xds = _open_dataset(
#             os.path.join(ms_store, n), load=load, file_system=file_system
#         )
#         # Skip empty tables
#         if not xds.coords and not xds.data_vars:
#             continue
#         if n in field_dict.keys():
#             field_and_source_xds_dict[field_dict[n]] = xds
#         else:
#             sub_xds_dict[n] = xds

#     return sub_xds_dict, field_and_source_xds_dict


# def _get_data_name(xds, data_group):
#     if "visibility" in xds.attrs["data_groups"][data_group]:
#         data_name = xds.attrs["data_groups"][data_group]["visibility"]
#     elif "spectrum" in xds.attrs["data_groups"][data_group]:
#         data_name = xds.attrs["data_groups"][data_group]["spectrum"]
#     else:
#         error_message = (
#             "No Visibility or Spectrum data variable found in data_group "
#             + data_group
#             + "."
#         )
#         logger.exception(error_message)
#         raise ValueError(error_message)
#     return data_name



from __future__ import annotations

from typing import Literal, Optional, Union, List
from xradio.schema.bases import (
    xarray_dataset_schema,
    xarray_dataarray_schema,
    dict_schema,
)
from xradio.schema.typing import Attr, Coord, Coordof, Data, Dataof, Name
import numpy

# Dimensions
Time = Literal["time"]
""" Observation time dimension """
Frequency = Literal["frequency"]
""" Frequency dimension """
FrequencySystemCal = Literal["frequency_system_cal"]
""" Frequency dimension in the system calibration dataset """
Polarization = Literal["polarization"]
""" Polarization dimension """
UvwLabel = Literal["uvw_label"]
""" Coordinate dimension of UVW data (typically shape 3 for 'u', 'v', 'w') """
SkyDirLabel = Literal["sky_dir_label"]
""" Coordinate labels of sky directions (typically shape 2 and 'ra', 'dec') """
LocalSkyDirLabel = Literal["local_sky_dir_label"]
""" Coordinate labels of local sky directions (typically shape 2 and 'az', 'alt') """
SphericalDirLabel = Literal["spherical_dir_label"]
""" Coordinate labels of spherical directions (shape 2 and 'lon', 'lat1' """
SkyPosLabel = Literal["sky_pos_label"]
""" Coordinate labels of sky positions (typically shape 3 and 'ra', 'dec', 'dist') """
SphericalPosLabel = Literal["spherical_pos_label"]
""" Coordinate labels of spherical positions (shape shape 3 and 'lon', 'lat1', 'dist2') """
EllipsoidPosLabel = Literal["ellipsoid_pos_label"]
""" Coordinate labels of geodetic earth location data (typically shape 3 and 'lon', 'lat', 'height')"""
CartesianPosLabel = Literal["cartesian_pos_label"]
""" Coordinate labels of geocentric earth location data (typically shape 3 and 'x', 'y', 'z')"""
nPolynomial = Literal["n_polynomial"]
""" For data that is represented as variable in time using Taylor expansion """
PolyTerm = Literal["poly_term"]
""" Polynomial term used in VLBI GAIN_CURVE """
LineLabel = Literal["line_label"]
""" Line labels (for line names and variables). """
FieldName = Literal["field_name"]
""" Field names dimension. """



import xarray as xr
from xradio._utils.schema import get_data_group_keys
from collections.abc import Mapping, Iterable
from typing import Any, Union


def get_data_group_name(
    xdx: Union[xr.Dataset, xr.DataTree], data_group_name: str = None
) -> str:

    if data_group_name is None:
        if "base" in xdx.attrs["data_groups"]:
            data_group_name = "base"
        else:
            data_group_name = list(xdx.attrs["data_groups"].keys())[0]

    return data_group_name


def create_new_data_group(
    xdx: Union[xr.Dataset, xr.DataTree],
    schema_name: str,
    new_data_group_name: str,
    data_group: dict,
    data_group_dv_shared_with: str = None,
) -> xr.DataTree:
    """Adds a data group to Xarray Data Structure (Dataset or DataTree).

    Parameters
    ----------
    new_data_group_name : str
        The name of the new data group to add.
    data_group : dict
        A dictionary containing the data group variables and their attributes.
    data_group_dv_shared_with : str, optional
        The name of the data group to share data variables with, by default "base"

    Returns
    -------
    xr.DataTree
        MSv4 DataTree with the new group added
    """

    data_group_dv_shared_with = get_data_group_name(xdx, data_group_dv_shared_with)

    default_data_group = xdx.attrs["data_groups"][data_group_dv_shared_with]

    new_data_group = {}

    data_group_keys = get_data_group_keys(schema_name)

    for key, optional in data_group_keys.items():
        if key in data_group:
            new_data_group[key] = data_group[key]
        else:
            if key not in default_data_group and not optional:
                raise ValueError(
                    f"Data group key '{key}' is required but not provided and not present in shared data group '{data_group_dv_shared_with}'."
                )
            elif key in default_data_group:
                new_data_group[key] = default_data_group[key]

    return new_data_group_name, new_data_group



import builtins
import dataclasses
import typing
import inspect
import functools
import warnings

import xarray
import numpy

from xradio.schema import (
    metamodel,
    bases,
    xarray_dataclass_to_array_schema,
    xarray_dataclass_to_dataset_schema,
    xarray_dataclass_to_dict_schema,
)
from xradio.schema.dataclass import value_schema
from xradio.schema.metamodel import AttrSchemaRef, ValueSchema


@dataclasses.dataclass
class SchemaIssue:
    """
    Representation of an issue found in a schema check

    As schemas can be quite big, ``path`` can be used to precisely locate the
    source of the issue.
    """

    path: typing.List[typing.Tuple[str, str]]
    """Path to offending data item, using pairs of (entity type, entity name).
    Entity types can be ``data_var``, ``coord`` or ``attr``.

    Example: ``[('data_var', 'foo'), ('coord','bar'), ('attr', 'asd')]``
    refers to ``obj.data_vars['foo'].cords['bar'].attrs['asd']``
    """
    message: str
    """
    Explanation of the issue
    """
    found: typing.Optional[typing.Any] = None
    """
    What was found. Can be any type (type, dtype, value)
    """
    expected: typing.Optional[typing.List[typing.Any]] = None
    """
    List of expected values. Can be any type (type, dtype, value)
    """

    def path_str(self):
        strs = []
        for path, ix in self.path:
            if ix is None or ix == "":
                strs.append(path)
            else:
                strs.append(f"{path}['{ix}']")
        return ".".join(strs)

    def __repr__(self):
        err = f"Schema issue with {self.path_str()}: {self.message}"
        if self.expected is not None:
            options = " or ".join(repr(option) for option in self.expected)
            if self.found is not None:
                err += f" (expected: {options} found: {repr(self.found)})"
            else:
                err += f" (expected: {options})"
        return err


class SchemaIssues(Exception):
    """
    List of issues found in a schema check

    Can be thrown as an exception, so we can report on multiple schema issues
    in one go.
    """

    issues: list[SchemaIssue]
    """List of issues found"""

    def __init__(self, issues=None):
        if issues is None:
            self.issues = []
        elif isinstance(issues, SchemaIssues):
            self.issues = list(issues.issues)
        else:
            self.issues = list(issues)

    def at_path(self, elem: str, ix: typing.Optional[str] = None) -> "SchemaIssues":
        for issue in self.issues:
            issue.path.insert(0, (elem, ix))
        return self

    def add(self, issue: SchemaIssue):
        self.issues.append(issue)

    def __iadd__(self, other: "SchemaIssues"):
        self.issues += other.issues
        return self

    def __add__(self, other: "SchemaIssues"):
        new_issues = SchemaIssues(self)
        new_issues += other
        return new_issues

    def __len__(self):
        return len(self.issues)

    def __getitem__(self, ix):
        return self.issues[ix]

    def __str__(self):
        if not self.issues:
            return "No schema issues found"
        else:
            issues_string = "\n * ".join(repr(issue) for issue in self.issues)
            return f"\n * {issues_string}"

    def __repr__(self):
        return f"SchemaIssues({str(self)})"

    def expect(
        self, elem: typing.Optional[str] = None, ix: typing.Optional[str] = None
    ):
        """
        Raises this object if issues were found

        :param elem: If given, will be added to path
        :param ix: If given, will be added to path
        :raises: SchemaIssues
        """
        if elem is not None:
            self.at_path(elem, ix)

        # Hide this function in pytest tracebacks
        __tracebackhide__ = True
        if self.issues:
            raise self


def check_array(
    array: xarray.DataArray, schema: typing.Union[type, metamodel.ArraySchema]
) -> SchemaIssues:
    """
    Check whether an xarray DataArray conforms to a schema

    :param array: DataArray to check
    :param schema: Schema to check against
    :returns: :py:class:`SchemaIssues` found
    """

    # Check that this is actually a DataArray
    if not isinstance(array, xarray.DataArray):
        raise TypeError(
            f"check_array: Expected xarray.DataArray, but got {type(array)}!"
        )
    if bases.is_dataarray_schema(schema):
        schema = xarray_dataclass_to_array_schema(schema)
    if not isinstance(schema, metamodel.ArraySchema):
        raise TypeError(f"check_array: Expected ArraySchema, but got {type(schema)}!")

    # Check dimensions
    issues = check_dimensions(array.dims, schema.dimensions)

    # Check type
    issues += check_dtype(array.dtype, schema.dtypes)

    # Check attributes
    issues += check_attributes(array.attrs, schema.attributes)

    # Check coordinates
    issues += check_data_vars(array.coords, schema.coordinates, "coords")

    return issues


def check_dataset(
    dataset: xarray.Dataset,
    schema: typing.Union[type, metamodel.DatasetSchema],
    allow_superflous_dims: typing.Set[str] = frozenset(),
) -> SchemaIssues:
    """
    Check whether an xarray DataArray conforms to a schema

    :param array: DataArray to check
    :param schema: Schema to check against
    :returns: :py:class:`SchemaIssues` found
    """

    # Check that this is actually a Dataset
    if not isinstance(dataset, xarray.Dataset):
        raise TypeError(
            f"check_dataset: Expected xarray.Dataset, but got {type(dataset)}!"
        )
    if bases.is_dataset_schema(schema):
        schema = xarray_dataclass_to_dataset_schema(schema)
    if not isinstance(schema, metamodel.DatasetSchema):
        raise TypeError(
            f"check_dataset: Expected DatasetSchema, but got {type(schema)}!"
        )

    # Check dimensions. Order does not matter on datasets
    issues = check_dimensions(
        dataset.dims,
        schema.dimensions,
        check_order=False,
        allow_superflous=allow_superflous_dims,
    )

    # Check attributes
    issues += check_attributes(dataset.attrs, schema.attributes)

    # Check coordinates
    issues += check_data_vars(dataset.coords, schema.coordinates, "coords")

    # Check data variables
    issues += check_data_vars(dataset.data_vars, schema.data_vars, "data_vars")

    return issues


def check_dimensions(
    dims: [str],
    expected: [[str]],
    check_order: bool = True,
    allow_superflous: typing.Set[str] = frozenset(),
) -> SchemaIssues:
    """
    Check whether a dimension list conforms to a schema

    :param array: Dimension list to check
    :param schema: Expected possibilities for dimension list
    :param check_order: Whether to check order of dimensions
    :returns: :py:class:`SchemaIssues` found
    """

    # Find a dimension list that matches
    dims_set = set(dims)
    best = None
    best_diff = 0
    for exp_dims in expected:
        exp_dims_set = set(exp_dims)

        # No match? Continue, but take note of best match
        if exp_dims_set - allow_superflous != dims_set - allow_superflous:
            diff = len(dims_set.symmetric_difference(exp_dims_set))
            if best is None or diff < best_diff:
                best = exp_dims_set
                best_diff = diff
            continue

        # Check that the order matches
        if not check_order or tuple(exp_dims) == tuple(dims):
            return SchemaIssues()

        return SchemaIssues(
            [
                SchemaIssue(
                    path=[("dims", None)],
                    message="Dimensions are in the wrong order! Consider transpose()",
                    found=list(dims),
                    expected=[exp_dims],
                )
            ]
        )

    # Dimensionality not supported - try to give a helpful suggestion
    hint_remove = [f"'{hint}'" for hint in (dims_set - best) - allow_superflous]
    hint_add = [f"'{hint}'" for hint in best - dims_set]
    if hint_remove and hint_add:
        message = f"Unexpected coordinates, replace {','.join(hint_remove)} by {','.join(hint_add)}?"
    elif hint_remove:
        message = f"Superfluous coordinate {','.join(hint_remove)}?"
    elif hint_add:
        message = f"Missing dimension {','.join(hint_add)}!"
    else:
        message = f"Unexpected dimensions/coordinates!"
    return SchemaIssues(
        [
            SchemaIssue(
                path=[("dims", None)],
                message=message,
                found=list(dims),
                expected=list(expected),
            )
        ]
    )


def check_dtype(dtype: numpy.dtype, expected: [numpy.dtype]) -> SchemaIssues:
    """
    Check whether a numpy dtype conforms to a schema

    :param dtype: Numeric type to check
    :param schema: Expected possibilities for dtype
    :returns: :py:class:`SchemaIssues` found
    """

    for exp_dtype_str in expected:
        exp_dtype = numpy.dtype(exp_dtype_str)
        # If the expected dtype has no size (e.g. "U", a.k.a. a string of
        # arbitrary length), we don't check itemsize, only kind.
        if (
            dtype.kind == exp_dtype.kind
            and exp_dtype.itemsize == 0
            or exp_dtype == dtype
        ):
            return SchemaIssues()

    # Not sure there's anything more helpful that we can do here? Any special
    # cases worth considering?
    return SchemaIssues(
        [
            SchemaIssue(
                path=[("dtype", None)],
                message="Wrong numpy dtype",
                found=dtype.str,
                expected=list(expected),
            )
        ]
    )


def check_attributes(
    attrs: typing.Dict[str, typing.Any],
    attrs_schema: typing.List[metamodel.AttrSchemaRef],
    attr_kind: str = "attrs",
) -> SchemaIssues:
    """
    Check whether an attribute set conforms to a schema

    :param attrs: Dictionary of attributes
    :param attrs_schema: Expected schemas
    :returns: :py:class:`SchemaIssues` found
    """

    issues = SchemaIssues()
    for attr_schema in attrs_schema:
        # Attribute missing is equivalent to a value of "None" is
        # equivalent for the purpose of the check
        val = attrs.get(attr_schema.name)
        if val is None:
            if not attr_schema.optional:
                issues.add(
                    SchemaIssue(
                        path=[(attr_kind, attr_schema.name)],
                        message="Non-optional attribute is missing!",
                        found=None,
                        expected=[attr_schema.type],
                    )
                )
            continue

        # Check actual value
        issues += _check_value(val, attr_schema).at_path(attr_kind, attr_schema.name)

    # Extra attributes are always okay

    return issues


def check_data_vars(
    data_vars: typing.Dict[str, xarray.DataArray],
    data_vars_schema: typing.List[metamodel.ArraySchemaRef],
    data_var_kind: str,
) -> SchemaIssues:
    """
    Check whether a data variable set conforms to a schema

    As data variables are data arrays, this will recurse into checking the
    array schemas

    :param data_vars: Dictionary(-like) of data_varinates
    :param data_vars_schema: Expected schemas
    :param datavar_kind: Either 'coords' or 'data_vars'
    :returns: :py:class:`SchemaIssues` found
    """

    assert data_var_kind in ["coords", "data_vars"]

    issues = SchemaIssues()
    for data_var_schema in data_vars_schema:
        allow_multiple_versions = False
        for attr in data_var_schema.attributes:
            if hasattr(attr, "name"):
                if attr.name == "allow_multiple_versions":
                    allow_multiple_versions = attr.default

        data_vars_names = []
        if allow_multiple_versions:
            for data_var_name in data_vars:
                if data_var_schema.name in data_var_name:
                    data_vars_names.append(data_var_name)
        else:
            data_vars_names = [data_var_schema.name]

        if (len(data_vars_names) == 0) and ~data_var_schema.optional:
            data_vars_names = [data_var_schema.name]

        for data_var_name in data_vars_names:
            data_var = data_vars.get(data_var_name)

            if data_var is None:
                if not data_var_schema.optional:
                    if data_var_kind == "coords":
                        message = (
                            f"Required coordinate '{data_var_schema.name}' is missing!"
                        )
                    else:
                        message = (
                            f"Required data variable '{data_var_schema.name}' is missing "
                            f"(have {','.join(data_vars)})!"
                        )
                    issues.add(
                        SchemaIssue(
                            path=[(data_var_kind, data_var_schema.name)],
                            message=message,
                        )
                    )
                continue

            # Check array schema
            issues += check_array(data_var, data_var_schema).at_path(
                data_var_kind, data_var_schema.name
            )

    # Extra data_varinates / data variables are always okay

    return issues


def check_dict(
    dct: dict, schema: typing.Union[type, metamodel.DictSchema]
) -> SchemaIssues:
    """
    Check whether a dictionary conforms to a schema

    :param dct: Dictionary to check
    :param schema: Dictionary schema to check against
    :returns: :py:class:`SchemaIssues` found
    """

    # Check that this is actually a dictionary
    if not isinstance(dct, dict):
        raise TypeError(f"check_dict: Expected dictionary, but got {type(dct)}!")
    if bases.is_dict_schema(schema):
        schema = xarray_dataclass_to_dict_schema(schema)
    if not isinstance(schema, metamodel.DictSchema):
        raise TypeError(f"check_dict: Expected DictSchema, but got {type(schema)}!")

    # Check attributes
    return check_attributes(dct, schema.attributes, attr_kind="")


def _check_value(val: typing.Any, schema: metamodel.ValueSchema):
    """
    Check whether value satisfies annotation

    If the annotation is a data array or dataset schema, it will be checked.

    :param val: Value to check
    :param schema: Schema of value
    :returns: Schema issues
    """

    # Unspecified?
    if schema.type is None:
        return SchemaIssues()

    # Optional?
    if schema.optional and val is None:
        return SchemaIssues()

    # Is supposed to be a data array?
    if schema.type == "dataarray":
        # Attempt to convert dictionaries automatically
        if isinstance(val, dict):
            try:
                val = xarray.DataArray.from_dict(val)
            except ValueError as e:
                expected = [xarray.DataArray]
                if schema.optional:
                    expected.append(type(None))
                return SchemaIssues(
                    [
                        SchemaIssue(
                            path=[], message=str(e), expected=expected, found=type(val)
                        )
                    ]
                )
            except TypeError as e:
                expected = [xarray.DataArray]
                if schema.optional:
                    expected.append(type(None))
                return SchemaIssues(
                    [
                        SchemaIssue(
                            path=[], message=str(e), expected=expected, found=type(val)
                        )
                    ]
                )

        if isinstance(val, xarray.DataArray):
            return check_array(val, schema.array_schema)

        expected = [xarray.DataArray]
        if schema.optional:
            expected.append(type(None))
        return SchemaIssues(
            [
                SchemaIssue(
                    path=[],
                    message=f"{type(val).__name__} is not an xarray.DataArray!",
                    expected=expected,
                    found=type(val),
                )
            ]
        )

    # Is supposed to be a dictionary?
    elif schema.type == "dict":
        if not isinstance(val, dict):
            expected = [dict]
            if schema.optional:
                expected.append(type(None))
            return SchemaIssues(
                [
                    SchemaIssue(
                        path=[],
                        message=f"{type(val).__name__} is not a dictionary!",
                        expected=expected,
                        found=type(val),
                    )
                ]
            )
        else:
            return check_dict(val, schema.dict_schema)

    elif schema.type == "list[str]":
        if not isinstance(val, list):
            expected = [list[str]]
            if schema.optional:
                expected.append(type(None))
            return SchemaIssues(
                [
                    SchemaIssue(
                        path=[],
                        message=f"{type(val).__name__} is not a list!",
                        expected=expected,
                        found=type(val),
                    )
                ]
            )
        if any(type(v) is not str for v in val):
            expected = [list[str]]
            if schema.optional:
                expected.append(type(None))
            return SchemaIssues(
                [
                    SchemaIssue(
                        path=[],
                        message=f"{type(val).__name__} is not a list of strings!",
                        expected=expected,
                        found=type(val),
                    )
                ]
            )
    elif schema.type in ["bool", "str", "int", "float"]:
        type_to_check = getattr(builtins, schema.type)
        if not isinstance(val, type_to_check):
            expected = [type_to_check]
            if schema.optional:
                expected.append(type(None))
            return SchemaIssues(
                [
                    SchemaIssue(
                        path=[],
                        message=f"{type(val).__name__} is not a {schema.type}!",
                        expected=expected,
                        found=type(val),
                    )
                ]
            )
    else:
        raise ValueError(f"Invalid typ_name in schema: {schema.type}")

    # List of literals given?
    if schema.literal is not None:
        for lit in schema.literal:
            if val == lit:
                return SchemaIssues()
        return SchemaIssues(
            [
                SchemaIssue(
                    path=[],
                    message=f"Disallowed literal value!",
                    expected=schema.literal,
                    found=val,
                )
            ]
        )

    return SchemaIssues()


_DATASET_TYPES = {}


def register_dataset_type(schema: metamodel.DatasetSchema):
    """
    Registers the given schema for usage with :py:meth:`check_datatree`

    This looks for a ``type`` attribute in the dataset schema, which
    must have a :py:class:`typing.Literal` type annotation specifying
    the type name of the dataset

    :param schema: Schema to register
    """

    # Find type attribute
    for attr in schema.attributes:
        if attr.name != "type":
            continue

        # Type should be a kind of literal
        if attr.literal is None:
            warnings.warn(
                f"In dataset schema {schema.schema_name}:"
                'Attribute "type" should be a literal!'
            )
            continue

        # Register type names
        for typ in attr.literal:
            assert isinstance(typ, str), (
                f"In dataset schema {schema.schema_name}:"
                'Attribute "type" should be a literal giving '
                "names of schema!"
            )
            _DATASET_TYPES[typ] = schema


def check_datatree(
    datatree: xarray.DataTree,
):
    """
    Check datatree for schema conformance

    This is the case if each contained :py:class:`~xarray.Dataset` conforms to a schema
    registed with Xradio.
    This works by looking for a ``type`` attribute in the :py:class:`~xarray.Dataset`, which
    must have a :py:class:`typing.Literal` type annotation specifying
    the name of the dataset schema.

    :param datatree: Data to check for schema conformance
    """

    # Loop through all groups in datatree
    issues = SchemaIssues()
    for xds_name in datatree.groups:
        # Ignore any leaf without data
        node = datatree[xds_name]
        if not node.has_data:
            continue

        # Look up schema
        typ = node.attrs.get("type")
        schema = _DATASET_TYPES.get(typ)
        if schema is None:
            issues.add(
                SchemaIssue(
                    [("", xds_name)],
                    message="Unknown dataset type!",
                    found=typ,
                    expected=None,
                )
            )
            continue

        # Determine dimensions inherited from parent
        # (they might show up as "superflous" for the child schema)
        parent_dims = frozenset()
        if node.parent is not None:
            parent_dims = set(node.parent.dims)

        # Check schema
        issues += check_dataset(node.dataset, schema, parent_dims).at_path("", xds_name)

    return issues


def schema_checked(fn, check_parameters: bool = True, check_return: bool = True):
    """
    Function decorator to check parameters and return value for
    schema conformance

    :param fn: Function to decorate
    :param check_parameters: Whether to check parameters. Can also
       pass an iterable with parameters to check
    :param check_return: Whether to check return value
    :returns: Decorated function
    """
    anns = inspect.getfullargspec(fn).annotations
    signature = inspect.signature(fn)

    if isinstance(check_parameters, bool):
        if check_parameters:
            parameters_to_check = set(signature.parameters)
        else:
            parameters_to_check = {}
    else:
        parameters_to_check = set(check_parameters)
        assert parameters_to_check.issubset(signature.parameters)

    @functools.wraps(fn)
    def _check_fn(*args, **kwargs):
        # Hide this function in pytest tracebacks
        # __tracebackhide__ = True

        # Bind parameters, collect (potential) issues
        bound = signature.bind(*args, **kwargs)
        issues = SchemaIssues()
        for arg, val in bound.arguments.items():
            if arg not in parameters_to_check:
                continue

            # Get annotation
            vschema = value_schema(anns.get(arg), "function", arg)
            pseudo_attr_schema = AttrSchemaRef(
                name=arg,
                **{
                    fld.name: getattr(vschema, fld.name)
                    for fld in dataclasses.fields(ValueSchema)
                },
            )
            issues += _check_value(val, pseudo_attr_schema).at_path(arg)

        # Any issues found? raise
        issues.expect()

        # Execute function
        result = fn(*args, **kwargs)

        # Check return
        if check_return:
            vschema = value_schema(anns.get("return"), "function", "return")
            pseudo_attr_schema = AttrSchemaRef(
                name="return",
                **{
                    fld.name: getattr(vschema, fld.name)
                    for fld in dataclasses.fields(ValueSchema)
                },
            )
            issues = _check_value(result, pseudo_attr_schema)
            issues.at_path("return").expect()

        return result

    return _check_fn



#################################
# Helper File
#
# Not exposed in API
#
#################################
import os
import warnings
from typing import Union

import xarray as xr
import re
import dask.array as da
from xradio._utils.schema import get_data_group_keys

try:
    from casacore import tables
except ImportError:
    import xradio._utils._casacore.casacore_from_casatools as tables


from xradio.image._util._casacore.xds_from_casacore import (
    _add_mask,
    _add_sky_or_aperture,
    _casa_image_to_xds_attrs,
    _casa_image_to_xds_coords,
    _get_mask_names,
    _get_persistent_block,
    _get_starts_shapes_slices,
    _get_transpose_list,
    _get_beam,
    _read_image_array,
)
from xradio.image._util._casacore.xds_to_casacore import (
    _coord_dict_from_xds,
    _history_from_xds,
    _imageinfo_dict_from_xds,
    _write_casa_data,
)
from xradio.image._util.common import (
    _aperture_or_sky,
    _get_xds_dim_order,
    _dask_arrayize_dv,
)
from xradio.image._util._casacore.common import _beam_fit_params, _open_image_ro

warnings.filterwarnings("ignore", category=FutureWarning)


def _squeeze_if_needed(ary: da, image_type: str) -> da:
    if image_type.upper() == "VISIBILITY_NORMALIZATION":
        shape = ary.shape
        if shape[3] != 1 or shape[4] != 1:
            raise ValueError(
                "VISIBILITY_NORMALIZATION casa image must have l and m of length 1. Found "
                + [shape[3], shape[4]]
            )
        ary = ary.squeeze(axis=(3, 4))
    return ary


def _get_casa_image_metadata(infile: str, do_sky_coords: bool, image_type: str) -> dict:
    image_full_path = os.path.expanduser(infile)
    with _open_image_ro(image_full_path) as casa_image:
        coords = casa_image.coordinates()
        cshape = casa_image.shape()
    ret = _casa_image_to_xds_coords(image_full_path, False, do_sky_coords, image_type)
    xds = ret["xds"]
    sphr_dims = ret["sphr_dims"]
    nchan = ret["xds"].sizes["frequency"]
    npol = ret["xds"].sizes["polarization"]
    dimorder = _get_xds_dim_order(ret["sphr_dims"], image_type)
    metadata = {
        "coords": coords,
        "cshape": cshape,
        "image_full_path": image_full_path,
        "nchan": nchan,
        "npol": npol,
        "dimorder": dimorder,
        # "xds_attrs": _casa_image_to_xds_attrs(image_full_path),
        "sphr_dims": sphr_dims,
        "xds": xds,
    }
    return metadata


def _load_casa_image_block(
    infile: str, block_des: dict, do_sky_coords: bool, image_type: str
) -> xr.Dataset:
    md = _get_casa_image_metadata(infile, do_sky_coords, image_type)
    coords = md["coords"]
    cshape = md["cshape"]
    dimorder = md["dimorder"]
    sphr_dims = md["sphr_dims"]
    nchan = md["nchan"]
    npol = md["npol"]
    xds = md["xds"].isel(block_des)
    image_full_path = md["image_full_path"]
    starts, shapes, slices = _get_starts_shapes_slices(block_des, coords, cshape)
    transpose_list, new_axes = _get_transpose_list(coords)
    block = _get_persistent_block(
        image_full_path, shapes, starts, dimorder, transpose_list, new_axes
    )
    block = _squeeze_if_needed(block, image_type)
    xds = _add_sky_or_aperture(
        xds, block, dimorder, image_full_path, sphr_dims, False, image_type
    )
    mymasks = _get_mask_names(image_full_path)
    for m in mymasks:
        full_path = os.sep.join([image_full_path, m])
        block = _get_persistent_block(
            full_path, shapes, starts, dimorder, transpose_list, new_axes
        )
        # data vars are all caps by convention
        mask_name = re.sub(r"\bMASK(\d+)\b", r"MASK_\1", m.upper())
        xds = _add_mask(xds, mask_name, block, dimorder)
    xds.attrs = _casa_image_to_xds_attrs(image_full_path)
    beam = _get_beam(image_full_path, nchan, npol, False, image_type)

    if beam is not None:
        selectors = {
            k: block_des[k]
            for k in ("time", "frequency", "polarization")
            if k in block_des
        }
        xds["BEAM_FIT_PARAMS_" + image_type.upper()] = beam.isel(selectors)
        xds["BEAM_FIT_PARAMS_" + image_type.upper()].attrs["type"] = (
            "beam_fit_params_" + image_type.lower()
        )
        xds[image_type.upper()].attrs[_beam_fit_params] = (
            "BEAM_FIT_PARAMS_" + image_type.upper()
        )
    return xds


def _open_casa_image(
    infile: str,
    chunks: Union[list, dict],
    verbose: bool,
    do_sky_coords: bool,
    masks: bool = True,
    history: bool = False,
    image_type: str = "SKY",
) -> xr.Dataset:
    md = _get_casa_image_metadata(infile, do_sky_coords, image_type)
    xds = md["xds"]
    dimorder = md["dimorder"]
    sphr_dims = md["sphr_dims"]
    img_full_path = md["image_full_path"]
    ary = _read_image_array(img_full_path, chunks, verbose=verbose)
    ary = _squeeze_if_needed(ary, image_type)
    xds = _add_sky_or_aperture(
        xds,
        ary,
        dimorder,
        img_full_path,
        sphr_dims,
        history,
        image_type,
    )
    if masks:
        mymasks = _get_mask_names(img_full_path)
        for m in mymasks:
            ary = _read_image_array(img_full_path, chunks, mask=m, verbose=verbose)
            # data var names are all caps by convention
            mask_name = re.sub(r"\bMASK(\d+)\b", r"MASK_\1", m.upper())
            xds = _add_mask(xds, mask_name, ary, dimorder)
    xds.attrs = _casa_image_to_xds_attrs(img_full_path)
    beam = _get_beam(
        img_full_path,
        xds.sizes["frequency"],
        xds.sizes["polarization"],
        True,
        image_type,
    )
    if beam is not None:
        xds["BEAM_FIT_PARAMS_" + image_type.upper()] = beam
        xds["BEAM_FIT_PARAMS_" + image_type.upper()].attrs["type"] = (
            "beam_fit_params_" + image_type.lower()
        )
        xds[image_type.upper()].attrs[_beam_fit_params] = (
            "BEAM_FIT_PARAMS_" + image_type.upper()
        )

    # xds = _add_coord_attrs(xds, ret["icoords"], ret["dir_axes"])
    xds = _dask_arrayize_dv(xds)

    return xds


def _xds_to_multiple_casa_images(xds: xr.Dataset, image_store_name: str) -> None:
    """Function disentagles xradio xr.Dataset into multiple casa images based on data_groups attribute.
    An xr.Dataset may contain multiple images (sky, residual, psf, etc) stored under different data variables sharing common coordinates.
    An addtional complication is that CASA images allow for internal masks and beam fit parameters to be stored alongside the main image data so these also need to be handled.
    This function creates separate casa images for each image type found in the data_groups attribute of the xr.Dataset.

    Parameters
    ----------
    xds : xr.Dataset
        The xradio xr.Dataset containing multiple images and associated data.
    image_store_name : str
        The base name or path for storing the output CASA images.
        If only one image is written, it will be named image_store_name, esle the images
        will be named image_store_name.<image_type> where <image_type> is sky, residual, point_spread_function, etc.
    """

    data_vars_name_set = set(xds.data_vars.keys())

    data_group_keys = list(get_data_group_keys(schema_name="image").keys())
    internal_image_types_to_exclude = [
        "flag",
        "beam_fit_params_sky",
        "beam_fit_params_point_spread_function",
    ]
    n_image_written = 0
    last_image_written = ""
    for data_group in xds.attrs["data_groups"].keys():
        for image_type in data_group_keys:
            if (image_type in xds.attrs["data_groups"][data_group]) and (
                image_type not in internal_image_types_to_exclude
            ):
                image_name = xds.attrs["data_groups"][data_group][image_type]
                if image_name in data_vars_name_set:
                    image_to_write_xds = xr.Dataset()
                    image_to_write_xds.attrs = xds.attrs.copy()

                    if image_type == "aperture":
                        image_to_write_xds["APERTURE"] = xds[image_name]
                    else:
                        image_to_write_xds["SKY"] = xds[image_name]

                    # This code handles adding internal masks and beam fit params if they exist.
                    if image_type == "sky":
                        if (
                            "beam_fit_params_sky"
                            in xds.attrs["data_groups"][data_group]
                        ):
                            beam_fit_params_name = xds.attrs["data_groups"][data_group][
                                "beam_fit_params_sky"
                            ]
                            image_to_write_xds["BEAM_FIT_PARAMS"] = xds[
                                beam_fit_params_name
                            ]

                        if "flag" in xds.attrs["data_groups"][data_group]:
                            mask_sky_name = xds.attrs["data_groups"][data_group]["flag"]
                            image_to_write_xds["MASK_0"] = xds[mask_sky_name]
                            image_to_write_xds["SKY"].attrs["flag"] = "MASK_0"

                    if image_type == "point_spread_function":
                        if (
                            "beam_fit_params_point_spread_function"
                            in xds.attrs["data_groups"][data_group]
                        ):
                            beam_fit_params_name = xds.attrs["data_groups"][data_group][
                                "beam_fit_params_point_spread_function"
                            ]
                            image_to_write_xds["BEAM_FIT_PARAMS"] = xds[
                                beam_fit_params_name
                            ]
                    outname = image_store_name + "." + image_type
                    _xds_to_casa_image(image_to_write_xds, outname)
                    if not os.path.exists(outname):
                        raise IOError(f"Failed to write CASA image {outname}")
                    n_image_written += 1
                    last_image_written = outname
                    data_vars_name_set.remove(image_name)
    if n_image_written == 0:
        raise ValueError("No valid image types found in xds to write to CASA images.")
    if n_image_written == 1:
        # rename the single written image to what the user requested
        os.rename(last_image_written, image_store_name)


def _xds_to_casa_image(xds: xr.Dataset, image_store_name: str) -> None:
    image_full_path = os.path.expanduser(image_store_name)
    _write_casa_data(xds, image_full_path)
    # create coordinates
    ap_sky = _aperture_or_sky(xds)
    coord = _coord_dict_from_xds(xds)
    ii = _imageinfo_dict_from_xds(xds)
    units = xds[ap_sky].attrs["units"] if "units" in xds[ap_sky].attrs else None
    miscinfo = (
        xds.attrs["user"]
        if "user" in xds.attrs and len(xds.attrs["user"]) > 0
        else None
    )
    tb = tables.table(
        image_full_path,
        readonly=False,
        lockoptions={"option": "permanentwait"},
        ack=False,
    )

    tb.putkeyword("coords", coord)
    tb.putkeyword("imageinfo", ii)
    if units:
        tb.putkeyword("units", units)
    if miscinfo:
        tb.putkeyword("miscinfo", miscinfo)
    tb.done()
    # history
    _history_from_xds(xds, image_full_path)



"""
Checks that an MSv4 converted from a generated MSv2 (gen_test_ms) has:
 - the expected sizes
 - input IDs (for antennas, fields, scans, etc.)
 - data column
 - etc. structure
of the generated MSv2 given as input to the MSv2=>MSv4 converter.
"""

import numpy as np
import xarray as xr


def check_msv4_matches_descr(msv4_xdt, msv2_descr):
    """
    Checks a single partition / MSv4

    Parameters
    ----------
    msv4_xdt : xr.DataTree
        An MSv4 DataTree containing one MSv4 partition

    msv2_descr : dict
        MSv2 description that was used to generate a test input MSv2 in gen_test_ms

    """
    msv2_to_msv4_data_colname = {
        "DATA": "VISIBILITY",
        "MODEL_DATA": "VISIBILITY_MODEL",
        "CORRECTED_DATA": "VISIBILITY_CORRECTED",
    }

    msv4_xds = msv4_xdt.ds
    assert (msv2_descr["nrows_per_ddi"],) == msv4_xds.time.shape
    nantennas = len(msv2_descr["ANTENNA"])
    nbaselines = int(nantennas * (nantennas - 1) / 2)
    assert (nbaselines,) == msv4_xds.baseline_id.shape
    assert (msv2_descr["nchans"],) == msv4_xds.frequency.shape
    assert (msv2_descr["nchans"],) == msv4_xds.frequency.shape
    assert (msv2_descr["npols"],) == msv4_xds.polarization.shape

    assert "scan_name" in msv4_xds
    assert len(np.unique(msv4_xds.coords["scan_name"])) == 1

    for msv2_data_colname in msv2_descr["data_cols"]:
        msv4_data_colname = msv2_to_msv4_data_colname[msv2_data_colname]
        assert msv4_data_colname in msv4_xds.data_vars

        data_shape = msv4_xds.data_vars[msv4_data_colname].shape
        assert msv2_descr["nrows_per_ddi"] == data_shape[0]
        assert nbaselines == data_shape[1]
        assert msv2_descr["nchans"] == data_shape[2]
        assert msv2_descr["npols"] == data_shape[3]

    ant_xds = msv4_xdt["antenna_xds"].ds
    assert "antenna_name" in ant_xds
    assert len(ant_xds.coords["antenna_name"]) == nantennas

    if msv2_descr["params"]["misbehave"]:
        expected_type = "field_and_source"
    else:
        expected_type = "field_and_source_ephemeris"

    field_and_source_name = "field_and_source_base_xds"
    assert (
        field_and_source_name
        == msv4_xdt.ds.attrs["data_groups"]["base"]["field_and_source"]
    )
    assert field_and_source_name in msv4_xdt
    assert msv4_xdt[field_and_source_name].ds.attrs["type"] == expected_type

    if msv2_descr["params"]["opt_tables"]:
        assert "system_calibration_xds" in msv4_xdt
        assert "weather_xds" in msv4_xdt
        if not msv2_descr["params"]["misbehave"]:
            assert "SOURCE_DIRECTION" in msv4_xdt["field_and_source_base_xds"].ds
            assert (
                msv4_xdt["field_and_source_base_xds"].attrs["type"]
                == "field_and_source_ephemeris"
            )

    if msv2_descr["params"]["vlbi_tables"]:
        assert "gain_curve_xds" in msv4_xdt
        assert "phase_calibration_xds" in msv4_xdt

    partition_info = msv4_xdt.xr_ms.get_partition_info()
    processor_info = msv4_xdt.ds.attrs["processor_info"]
    if msv2_descr["params"]["misbehave"]:
        # SPW names should be empty string in MSv2
        assert partition_info["spectral_window_name"] == "spw_0"
        assert not processor_info["type"]
        assert not processor_info["sub_type"]
    else:
        assert partition_info["spectral_window_name"]
        assert processor_info["type"]
        assert processor_info["sub_type"]

    observation_info = msv4_xdt.ds.attrs["observation_info"]
    if msv2_descr["params"]["opt_tables"] and not msv2_descr["params"]["misbehave"]:
        assert "session_reference_UID" in observation_info
    else:
        assert "session_reference_UID" not in observation_info

    if (
        (msv2_descr["params"]["opt_tables"] and not msv2_descr["params"]["misbehave"])
    ) or "OBSERVATION" in msv2_descr:
        assert "execution_block_UID" in observation_info
        assert "scheduling_block_UID" in observation_info
    else:
        assert "execution_block_UID" not in observation_info
        assert "scheduling_block_UID" not in observation_info


def check_processing_set_matches_msv2_descr(
    proc_set_xdt: xr.DataTree, msv2_descr: dict
):
    """
    Parameters
    ----------
    proc_set_xdt : xr.DataTree
        A processing set DataTree containing MSv4s to be checked

    msv2_descr : dict
        MSv2 description that was used to generate a test input MSv2 in gen_test_ms

    Returns
    -------

    """
    for _msv4_name, msv4_xdt in proc_set_xdt.items():
        check_msv4_matches_descr(msv4_xdt, msv2_descr)
